import re
import time
from bloonix.logger import logger

class ArchiveStats:
    def __init__(self, pm, queue, db, ds, config):
        self.pm = pm
        self.queue = queue
        self.db = db
        self.ds = ds
        self.config = config
        self.log = logger("bloonix")
        self.log.set_pattern("%Z", "archive_stats")

    def run(self):
        while self.pm.done is False:
            try:
                self.read_queue()
            except Exception:
                self.log.exception()
                self.ds.disconnect()
            time.sleep(1)

    def read_queue(self):
        while self.pm.done is False:
            self.pm.is_waiting()
            service_id = self.queue.pull(QUEUE_ARCHIVE_STATS)
            if not service_id or not re.search(r"^\d+\Z", service_id):
                break
            self.pm.is_processing()
            self.db.connect()
            self.ds.connect()
            self.process(int(service_id))
            self.db.disconnect()
            self.ds.disconnect()

    def process(self, service_id):
        history = self.db.get_sys_options().get("stat_hourly_history", 120)
        year = time.localtime().tm_year
        month = time.localtime().tm_mon
        for i in range(history):
            month -=1
            if month == 0:
                month = 12
                year -= 1
        min_start_time = utils.unixtime("%s-%02d-01 00:00:00" % (year, month))

        # Don't put load on today's partition. 1d + 1h = 90000
        end_time = self.calc_hour(int(time.time()) - 90000)
        self.log.notice("start processing service", service_id)
        runtime = time.time()
        processed_rows = 0
        insert_data = []
        inserted_rows = 0
        start_time = None

        max_time = self.ds.get_max_time_by_service(service_id)
        if max_time is None:
            min_time = self.ds.get_min_time_by_service(service_id)
            if min_time is not None:
                start_time = min_time
                self.log.info(
                    "service", service_id,
                    "has no data, start at",
                    utils.timestamp(start_time),
                    "(%s)" % start_time)
        else:
            start_time = max_time

        if start_time is None:
            return

        start_time = self.calc_hour(start_time)
        if start_time < min_start_time:
            start_time = min_start_time

        if start_time + 3600 > end_time:
            return

        self.log.info(
            "process service", service_id,
            "period", utils.timestamp(start_time),
            "-", utils.timestamp(end_time))

        while start_time <= end_time:
            from_time = start_time
            start_time = to_time = start_time + 3600
            self.ds.create_archive_partition(utils.timestamp(from_time))
            self.log.debug(
                "get stat for service", service_id,
                utils.timestamp(from_time), "-", utils.timestamp(to_time))

            data = {}
            for row in self.ds.get_stats_for_archive(service_id, from_time, to_time):
                processed_rows += 1
                item_id = row["item_id"]
                if item_id not in data:
                    data[item_id] = {
                        "time": from_time,
                        "service_id": service_id,
                        "item_id": item_id,
                        "avg": row["value"],
                        "min": row["value"],
                        "max": row["value"],
                        "count": 1,
                    }
                    continue

                item = data[item_id]
                if row["value"] < item["min"]:
                    item["min"] = row["value"]
                if row["value"] > item["max"]:
                    item["max"] = row["value"]
                item["avg"] += row["value"]
                item["count"] += 1

            for item_id, item in data.items():
                item["avg"] = float(format(item["avg"] / item["count"], ".3f"))
                del item["count"]
                insert_data.append(item)

            if len(insert_data) >= 100:
                inserted_rows += len(insert_data)
                self.ds.create_archive_data(insert_data)
                insert_data.clear()

        if len(insert_data):
            inserted_rows += len(insert_data)
            self.ds.create_archive_data(insert_data)
            insert_data.clear()

        runtime = time.time() - runtime
        self.log.notice(
            "service", service_id,
            "processed=%s" % processed_rows,
            "inserted=%s" % inserted_rows,
            "runtime=%.2fs" % runtime)

    def calc_hour(self, t):
        return t - (t % 3600)
