import re
import time
from bloonix.dbi import DBI
from bloonix.sql import SQL
from bloonix.logger import logger
from bloonix import utils
from bloonix import mstatus

class Database(DBPartition):
    def __init__(self, **kwargs):
        self.log = logger("bloonix")
        self.dbi = DBI.setup(**kwargs)
        self.do = self.dbi.do
        self.ping = self.dbi.ping
        self.sql = SQL(driver=kwargs["driver"])

    def get_host_by_auth(self, host, password, location_id):
        if location_id is None:
            return self.__get_host_by_host_secret_auth(host, password)

        return self.__get_host_by_location_group_auth(host, password, location_id)

    def __get_host_by_host_secret_auth(self, host, password):
        if re.search(r"^\d+\Z", host):
            column = "host_secret.host_id"
        else:
            column = "host.hostname"

        stmt, bind = self.sql.select(
            table=[
                "host", "*"
            ],
            join=[
                "inner", ["host_secret", "host.id", "host_secret.host_id"]
            ],
            condition=[
                "where", [column, host],
                "and", ["host_secret.password", password]
            ]
        )
        return self.dbi.fetchone(stmt, bind)

    def get_host(self, host):
        if isinstance(host, int) or re.search(r"^\d+\Z", host):
            column = "id"
        else:
            column = "hostname"
        stmt, bind = self.sql.select(
            table="host",
            columns="*",
            condition=["where", [column, host]])
        return self.dbi.fetchone(stmt, bind)

    def get_service(self, service_id):
        stmt, bind = self.sql.select(
            table="service",
            columns="host_id",
            condition=["where", ["id", service_id]])
        return self.dbi.fetchone(stmt, bind)

    def get_sys_options(self):
        stmt, bind = self.sql.select(
            table="sys_options",
            columns="*",
            condition=[
                "where", [
                    "key", "in", [
                        "event_history",
                        "stat_history",
                        "stat_hourly_history",
                        "result_history",
                        "notification_history",
                        "user_tracking_history",
                    ]
                ]
            ]
        )
        data = {}
        for row in self.dbi.fetchall(stmt, bind):
            key = row["key"]
            value = row["value"]
            data_type = row["data_type"]
            if data_type == "int":
                data[key] = int(value)
            elif data_type == "float":
                data[key] = float(value)
            elif data_type == "json":
                data[key] = json.loads(value)
            elif data_type == "string":
                data[key] = value
        return data

    def missing_services(self):
        stmt, bind = self.sql.select(
            table=[
                "service_parameter", "id AS service_parameter_id",
                "host_template", "host_id"
            ],
            join=[
                "inner", [
                    "host_template",
                    "service_parameter.template_id",
                    "host_template.template_id"
                ]
            ],
            condition=[
                "where not exists", {
                    "table": "service",
                    "columns": 1,
                    "condition": [
                        "where", ("service_parameter_id", "service_parameter.id"),
                        "and", ("host_id", "host_template.host_id")
                    ]
                }
            ]
        )
        return self.dbi.fetchall(stmt, bind)

    def create_service(self, data):
        stmt, bind = self.sql.insert(
            table="service",
            data=data)
        self.dbi.do(stmt, bind)

    def non_existent_services(self):
        stmt, bind = self.sql.select(
            table="service",
            columns="id",
            join=[
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id"
                ],
                "left outer", [
                    "host_template",
                    "service.host_id",
                    "host_template.host_id"
                ]
            ],
            condition=[
                "where", ["service_parameter.template_id", "is not", None],
                "and", ["host_template.template_id", "is", None]
            ]
        )
        return self.dbi.fetchall(stmt, bind)

    def delete_service(self, data):
        stmt, bind = self.sql.delete(
            table="service",
            condition=["where", ["id", data]])
        self.dbi.do(stmt, bind)

    def get_services(
            self, host_id=None, service_ids=None,
            with_vars=None, with_check_wtrm=None,
            remote_check=None):

        if not host_id and not service_ids:
            raise Exception("missing mandatory option host_id or service_id")

        condition = ["where"]

        if host_id:
            condition.append(["service.host_id", host_id])
            if with_check_wtrm is None or with_check_wtrm is False:
                condition.append("and")
                condition.append(["service_parameter.command", "!=", "check-wtrm"])
            if remote_check is not None:
                condition.append("and")
                condition.append(["service_parameter.remote_check", remote_check])
        elif service_ids:
            condition.append(["service.id", service_ids])

        stmt, bind = self.sql.select(
            table=[
                "service", "*",
                "service_parameter", [
                    "template_id",
                    "service_name",
                    "command",
                    "host_alive_check",
                    "command_options",
                    "satellite_type",
                    "satellites",
                    "description",
                    "interval",
                    "retry_interval",
                    "attempt_max",
                    "notification_interval",
                    "fd_enabled",
                    "fd_time_range",
                    "fd_flap_count",
                    "remote_check",
                ],
            ],
            join=[
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id"
                ],
            ],
            condition=condition
        )

        services = {}

        for service in self.dbi.fetchall(stmt, bind):
            services[service["id"]] = service

        if with_vars is True:
            self.add_template_vars(services)

        #self.log.warning("DEBUG")
        #self.log.warning("with_vars", with_vars)
        #self.log.warning("with_check_wtrm", with_check_wtrm)
        #self.log.warning("remote_check", remote_check)
        #self.log.warning(services, dump=True)
        return services

    def get_wtrm_services(self):
        stmt, bind = self.sql.select(
            table=[
                "service", [
                    "id",
                    "last_check",
                    "host_id",
                ],
                "service_parameter", [
                    "interval",
                    "retry_interval",
                ],
            ],
            join=[
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id"
                ],
            ],
            condition=[
                "where", ["service_parameter.command", "check-wtrm"],
                "and", ["service_parameter.remote_check", 1],
            ])
        return self.dbi.fetchall(stmt, bind)

    def add_template_vars(self, services):
        templates = {}

        for service_id, service in services.items():
            template_id = service["template_id"]

            if not template_id:
                continue

            if template_id not in templates:
                s, b = self.sql.select(
                    table="template",
                    columns="variables",
                    condition=["where", ["id", template_id]]
                )
                templates[template_id] = self.dbi.fetchone(s, b)

            # create a safe copy of all variables
            service["template_variables"] = dict(templates[template_id]["variables"])

    def get_location_group(self, location_id, password):
        stmt, bind = self.sql.select(
            table="location_group",
            columns="allow_from",
            condition=[
                "where", ["id", location_id],
                "and", ["password", password]
            ]
        )
        return self.dbi.fetchone(stmt, bind)

    def get_location_group_hosts(self, location_id):
        stmt, bind = self.sql.select(
            table="host_secret",
            columns=["host_id", "password"],
            join=[
                "inner", [
                    "host_location_group",
                    "host_secret.host_id",
                    "host_location_group.host_id"
                ]
            ],
            condition=[
                "where", [
                    "host_location_group.location_group_id", location_id
                ]
            ]
        )
        return self.dbi.fetchall(stmt, bind)

    def get_satellites(self, satellites, config):
        stmt, bind = self.sql.select(
            table="satellite",
            columns=["id", "name", "hostname", "port", "authkey"],
            condition=["where", ["id", "in", satellites]])
        satellites = {}
        for satellite in self.dbi.fetchall(stmt, bind):
            satellites[satellite["id"]] = satellite
            if satellite["authkey"] == "secret":
                if satellite["hostname"] not in config:
                    self.log.error(
                        "%s not in config[satellites]"
                        % satellite["hostname"])
                satellite["authkey"] = config.get(satellite["hostname"], "bloonix")
        return satellites

    def host_alive_critical(self, host_id, services):
        stmt, bind = self.sql.select(
            table="service",
            columns="id",
            join=[
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id"
                ]
            ],
            condition=[
                "where", ["service.host_id", host_id],
                "and", ["service_parameter.host_alive_check", 1],
                "and", ["service.status", mstatus.CRITICAL],
            ],
            limit=1
        )
        host_alive_checks = self.dbi.fetchall(stmt, bind)
        for service in host_alive_checks:
            if service["id"] in services:
                return False
        if len(host_alive_checks):
            return True
        return False

    def get_plugin_info(self, command):
        stmt, bind = self.sql.select(
            table="plugin",
            columns="*",
            condition=["where", ["command", command]])
        return self.dbi.fetchone(stmt, bind)

    def get_maintenance(self):
        stmt, bind = self.sql.select(
            table="sys_options",
            columns="*",
            condition=["where", ["key", "in", ["has_maintenance", "plugin_version"]]])
        rows = self.dbi.fetchall(stmt, bind)
        data = {
            rows[0]["key"]: rows[0]["value"],
            rows[1]["key"]: rows[1]["value"],
        }
        data["has_maintenance"] = int(data["has_maintenance"])
        return data

    def column_exists(self, table, column):
        stmt, bind = self.sql.select(
            table=table,
            columns=column,
            limit=1)
        try:
            self.dbi.fetchone(stmt, bind)
            return True
        except Exception:
            pass

    def get_host_downtimes(self, host_id):
        stmt, bind = self.sql.select(
            table="host_downtime",
            columns="*",
            condition=["where", ["host_id", host_id]])
        return self.dbi.fetchall(stmt, bind)

    def get_service_downtimes(self, services):
        stmt, bind = self.sql.select(
            table="service_downtime",
            columns="*",
            condition=["where", ["service_id", services]])
        return self.dbi.fetchall(stmt, bind)

    def update_service(self, service_id, data):
        stmt, bind = self.sql.update(
            table="service",
            data=data,
            condition=["where", ["id", service_id]])
        self.dbi.do(stmt, bind)

    def update_host(self, host_id, data):
        stmt, bind = self.sql.update(
            table="host",
            data=data,
            condition=["where", ["id", host_id]])
        self.dbi.do(stmt, bind)

    def create_event(self, data):
        stmt, bind = self.sql.insert(table="event", data=data)
        self.dbi.do(stmt, bind)

    def update_event(self, service_id, attempts):
        stmt, bind = self.sql.update(
            table="event",
            data={"attempts": attempts},
            condition=[
                "where", ["service_id", service_id],
                "and", ["time", "=", {
                    "table": "event",
                    "columns": "max(time)",
                    "condition": ["where", ["service_id", service_id]]
                }]
            ])
        self.dbi.do(stmt, bind)

    def update_last_notification(self, service_id, data):
        stmt, bind = self.sql.update(
            table="service",
            data=data,
            condition=["where", ["id", service_id]])
        self.dbi.do(stmt, bind)

    def get_dependencies(self, service_id):
        stmt, bind = self.sql.select(
            table=[
                "dependency", "*",
                "service", "status AS service_has_on_status"
            ],
            join=[
                "inner", [
                    "service",
                    "dependency.on_service_id",
                    "service.id"
                ]
            ],
            condition=[
                "where", ["dependency.service_id", service_id]
            ]
        )
        return self.dbi.fetchall(stmt, bind)

    def host_contact_ms(self, host_id):
        stmt, bind = self.sql.select(
            distinct=True,
            table="contact",
            columns="*",
            join=[
                "inner", [
                    "contact_contactgroup",
                    "contact.id",
                    "contact_contactgroup.contact_id"
                ],
                "inner", [
                    "host_contactgroup",
                    "contact_contactgroup.contactgroup_id",
                    "host_contactgroup.contactgroup_id"
                ]
            ],
            condition=[
                "where", ["host_contactgroup.host_id", host_id],
                "and", ["contact.enabled", 1]
            ])
        contacts = {}
        for contact in self.dbi.fetchall(stmt, bind):
            contact["message_services"] = self.message_service_by_contact(contact["id"])
            contacts[contact["id"]] = contact
        return contacts

    def service_contact_ms(self, service_ids, skip_contact_ids):
        if not len(skip_contact_ids):
            skip_contact_ids.append(0)

        stmt, bind = self.sql.select(
            table=[
                "contact", "*",
                "service_contactgroup", "service_id",
            ],
            join=[
                "inner", [
                    "contact_contactgroup",
                    "contact.id",
                    "contact_contactgroup.contact_id"
                ],
                "inner", [
                    "service_contactgroup",
                    "contact_contactgroup.contactgroup_id",
                    "service_contactgroup.contactgroup_id"
                ]
            ],
            condition=[
                "where", [
                    "service_contactgroup.service_id",
                    service_ids
                ],
                "and", [
                    "contact.enabled", 1
                ],
                "and", [
                    "contact_contactgroup.contact_id",
                    "not in",
                    skip_contact_ids
                ],
            ]
        )
        contacts = {}
        for contact in self.dbi.fetchall(stmt, bind):
            _id = contact["id"]
            if _id not in contacts:
                contacts[_id] = contact
                contacts[_id]["services"] = []
                contacts[_id]["message_services"] = self.message_service_by_contact(_id)
            contacts[_id]["services"].append(contact["service_id"])
            del contact["service_id"]
        return contacts

    def message_service_by_contact(self, contact_id):
        stmt, bind = self.sql.select(
            table=[
                "contact_message_service", [
                    "send_to",
                    "notification_level",
                    "timeperiods"
                ],
                "message_service", [
                    "builtin",
                    "script",
                    "name",
                    "params",
                    "message_template"
                ]
            ],
            join=[
                "inner", [
                    "message_service",
                    "contact_message_service.message_service_id",
                    "message_service.id"
                ]
            ],
            condition=[
                "where", ["contact_message_service.contact_id", contact_id],
                "and", ["contact_message_service.enabled", 1],
                "and", ["message_service.enabled", 1]
            ]
        )
        return self.dbi.fetchall(stmt, bind)

    def get_remote_services_without_wtrm(self):
        stmt, bind = self.sql.select(
            table=[
                "service", [
                    "id",
                    "host_id",
                    "last_check"
                ],
                "service_parameter", [
                    "interval",
                    "retry_interval",
                ]
            ],
            join=[
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id"
                ]
            ],
            condition=[
                "where", ["service_parameter.remote_check", 1],
                "and", ["service_parameter.command", "!=", "check-wtrm"],
            ]
        )
        return self.dbi.fetchall(stmt, bind)

    def get_timed_out_services(self, last_check_time):
        select = {
            "distinct": True,
            "table": [
                "service", [
                    "host_id",
                    "service_parameter_id",
                ],
                "service_parameter", [
                    "interval_changed_at",
                ],
            ],
            "join": [
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id",
                ],
            ],
            "condition": [
                "where", ["service.force_check", "!=", 2],
                "and", {
                    "math": [
                        "service.last_check",
                        "+",
                        "service_parameter.interval",
                        "<",
                        last_check_time
                    ]
                }
            ],
            "comment": "timed_out_services",
        }
        stmt, bind = self.sql.select(**select)
        return self.dbi.fetchall(stmt, bind)

    def fix_interval_changed_at(self, ids, t):
        stmt, bind = self.sql.update(
            table="service_parameter",
            data={"interval_changed_at": t},
            condition=["where", ["id", "in", ids]])
        self.dbi.do(stmt, bind)

    def delete_expired_downtimes(self):
        for table in ["host_downtime", "service_downtime"]:
            stmt, bind = self.sql.delete(
                table=table,
                condition=[
                    "where", ["end", "is not", None],
                    "and", ["end", "<", int(time.time()) - 43200],
                ])
            self.dbi.do(stmt, bind)

    def create_notification(self, data):
        stmt, bind = self.sql.insert(
            table="notification",
            data=data)
        self.dbi.do(stmt, bind)

    def delete_notifications(self, _time):
        stmt, bind = self.sql.delete(
            table="notification",
            condition=["where", ["time", "<", _time]])
        self.dbi.do(stmt, bind)

    def delete_user_tracking(self, _time):
        stmt, bind = self.sql.delete(
            table="user_tracking",
            condition=["where", ["time", "<", _time]])
        self.dbi.do(stmt, bind)

    def get_service_status(self, ids):
        services = {}
        stmt, bind = self.sql.select(
            table=["service", ["id", "status"]],
            condition=["where", ["id", "in", ids]])
        for service in self.dbi.fetchall(stmt, bind):
            services[service["id"]] = service["status"]
        return services

    def get_primary_keepalived(self):
        stmt, bind = self.sql.select(
            table="keepalived",
            columns="me",
            order=["asc", "id"])
        return self.dbi.fetchone(stmt, bind)

    def update_me_keepalived(self, me, _time):
        stmt, bind = self.sql.select(
            table="keepalived",
            columns="*",
            condition=["where", ["me", me]])
        if self.dbi.fetchone(stmt, bind):
            stmt, bind = self.sql.update(
                table="keepalived",
                data={"me": me, "time": _time},
                condition=["where", ["me", me]])
        else:
            stmt, bind = self.sql.insert(
                table="keepalived",
                data={"me": me, "time": _time})
        try:
            self.dbi.do(stmt, bind)
        except Exception:
            self.log.exception()

    def delete_keepalived(self, diff_time):
        stmt, bind = self.sql.delete(
            table="keepalived",
            condition=["where", ["time", "<", int(time.time()) - diff_time]])
        self.dbi.do(stmt, bind)
        stmt, bind = self.sql.delete(
            table="keepalived",
            condition=["where", ["time", ">", int(time.time()) + diff_time]])
        self.dbi.do(stmt, bind)

    def delete_me_keepalived(self, me):
        stmt, bind = self.sql.delete(
            table="keepalived",
            condition=["where", ["me", me]])
        self.dbi.do(stmt, bind)

    def get_number_of_plugin_metrics(self):
        plugins = {}
        stmt, bind = self.sql.select(
            table="plugin",
            columns=["command", "metrics"])
        for plugin in self.dbi.fetchall(stmt, bind):
            plugins[ plugin["command"] ] = len(plugin["metrics"])
        return plugins

    def get_services_for_archiver(self):
        plugins = self.get_number_of_plugin_metrics()
        services = []
        stmt, bind = self.sql.select(
            table=[
                "service", "id",
                "service_parameter", "command",
            ],
            join=[
                "inner", [
                    "service_parameter",
                    "service.service_parameter_id",
                    "service_parameter.id"
                ]
            ])
        for service in self.dbi.fetchall(stmt, bind):
            command = service["command"]
            if command in plugins and plugins[command] > 0:
                services.append(service["id"])
        return services

    def connect(self):
        return self.dbi.connect()

    def disconnect(self):
        return self.dbi.disconnect()
