import sys
import time
from bloonix import utils
from bloonix.logger import logger

class DBManager(CheckPrimary):
    def __init__(self, pm, queue, db, ds, config, me):
        self.pm = pm
        self.queue = queue
        self.db = db
        self.ds = ds
        self.config = config
        self.me = me
        self.log = logger("bloonix")
        self.log.set_pattern("%Z", "db_manager")
        self.db_lock = Database(**self.config["database"])
        self.db_lock.dbi.autocommit_default(False)
        start_time = int(time.time())
        self.schedule = [
            # check_missing_services and check_non_existent_services
            # should never come to a result when the templates, hosts
            # and services are created and deleted through the WebGUI.
            # If an eager user tries to create these things directly
            # in the database and does something wrong, the stuff
            # will be cleaned up here.
            ["create_partition", 0, 43200],
            ["delete_partition", 0, 43200],
            ["check_missing_services", 0, 43200],
            ["check_non_existent_services", 0, 43200],
            ["delete_expired_downtimes", 0, 600],
            ["delete_notifications", 0, 600],
            ["delete_user_tracking", 0, 600],
            ["archive_stats", start_time, 3600],
            ["delete_archive_partitions", 0, 43200],
        ]

    def run(self):
        while self.pm.done is False:
            try:
                self.db.connect()
                if self.am_i_primary():
                    self.ds.connect()
                    self._run()
            except Exception:
                self.log.exception()
            finally:
                self.db.disconnect()
                self.ds.disconnect()
            time.sleep(1)

    def _run(self):
        for s in self.schedule:
            if self.pm.done:
                break
            attr, last_check, interval = s
            # Fix time issues
            if last_check > time.time():
                s[1] = last_check = time.time()
            if last_check + interval < time.time():
                callback = getattr(self, attr)
                try:
                    ret = callback()
                    s[1] = time.time()
                except Exception as err:
                    self.log.error("the call of %s failed: %s" % (attr, err))
                    s[1] += 60 # try in a minute again

    def create_partition(self):
        self.db.create_partition(
            table="event",
            days=10)
        self.ds.create_partition(
            table="stat",
            days=10)
        self.ds.create_partition(
            table="result",
            days=10)

    def delete_partition(self):
        sys_options = self.db.get_sys_options()
        self.db.delete_partition(
            table="event",
            keep_days=sys_options["event_history"])
        self.ds.delete_partition(
            table="stat",
            keep_days=sys_options["stat_history"])
        self.ds.delete_partition(
            table="result",
            keep_days=sys_options["result_history"])

    def check_missing_services(self):
        self.log.notice("fetch missing services")
        lock_table = {"db_manager": "write"}
        try:
            self.db_lock.dbi.connect()
            self.db_lock.dbi.lock(lock_table)
            services = self.db.missing_services()
            if len(services):
                self.create_missing_services(services)
            else:
                self.log.notice("no services missing")
        except Exception:
            self.log.exception()
        try:
            self.db_lock.dbi.commit()
            self.db_lock.dbi.unlock(lock_table)
        except Exception:
            self.log.exception()
        self.db_lock.disconnect()
        return True

    def create_missing_services(self, services):
        total = len(services)
        count = 0
        self.log.notice("%s services missing" % total)
        while len(services):
            if self.pm.done:
                break

            bulk_data = []
            while len(services) and len(bulk_data) < 10000:
                bulk_data.append(services.pop())

            count += len(bulk_data)
            self.log.info(
                "create service %s/%s"
                % (count, total))
            time.sleep(0.02)
            self.db.create_service(bulk_data)
        self.log.notice("%s services created" % count)
        return True

    def check_non_existent_services(self):
        self.log.notice("fetch non existent services")
        services = self.db.non_existent_services()
        if len(services):
            total = len(services)
            count = 0
            self.log.notice("%s services does not exist any more" % total)

            while len(services):
                if self.pm.done:
                    break

                bulk_data = []
                while len(services) and len(bulk_data) < 10000:
                    s = services.pop()
                    bulk_data.append(s["id"])

                count += len(bulk_data)
                self.log.info(
                    "delete service %s/%s"
                    % (count, total))
                time.sleep(0.02)
                self.db.delete_service(bulk_data)
            self.log.notice("%s services deleted" % count)
        return True

    def delete_expired_downtimes(self):
        self.db.delete_expired_downtimes()

    def delete_notifications(self):
        sys_options = self.db.get_sys_options()
        self.db.delete_notifications(
            int(time.time()) - sys_options["notification_history"] * 86400)

    def delete_user_tracking(self):
        sys_options = self.db.get_sys_options()
        self.db.delete_notifications(
            int(time.time()) - sys_options["user_tracking_history"] * 86400)

    def archive_stats(self):
        self.log.notice("run archive manager")
        queue_len = self.queue.llen(QUEUE_ARCHIVE_STATS)
        if queue_len > 0:
            self.log.notice("skip archive manager, there are %s services in queue" % queue_len)
            return
        services = self.db.get_services_for_archiver()
        self.log.notice("archive manager: push", len(services), "services into queue")
        for service in services:
            self.queue.push(QUEUE_ARCHIVE_STATS, service)

    def delete_archive_partitions(self):
        self.ds.connect()
        self.db.connect()

        partitions = []
        history = self.db.get_sys_options().get("stat_hourly_history", 120)
        year = time.localtime().tm_year
        month = time.localtime().tm_mon
        for i in range(history):
            partitions.append("stat_hourly_p%d%02d" % (year, month))
            month -=1
            if month == 0:
                month = 12
                year -= 1

        for partition in self.ds.get_archive_partitions():
            if partition not in partitions:
                self.log.notice("delete partition %s" % partition)
                self.ds.delete_archive_partition(partition)
