import time

from bloonix.logger import logger
from bloonix import mstatus
from bloonix import netaddr

class Listener:
    def __init__(self, pm, sock, queue, db, ds, config):
        self.sock = sock
        self.queue = queue
        self.db = db
        self.ds = ds
        self.pm = pm
        self.config = config
        self.log = logger("bloonix")
        self.log.set_pattern("%Z", "listener")
        self.validate = Validate()

    def run(self):
        while self.pm.done is False:
            try:
                self.log.set_pattern("%X", "n/a") # client address
                self.log.set_pattern("%Y", "n/a") # host id
                self.pm.is_waiting()
                self.log.debug("waiting for new request")
                client = self.sock.accept()
                self.log.debug("received request")
                if client and self.pm.done is False:
                    self.pm.is_processing()
                    self.process_request(client)
            except Exception:
                # On SIG{INT|TERM} pm.done is set to True,
                # don't wanna see any error message if that
                # happends.
                if self.pm.done is False:
                    self.log.exception()
                    time.sleep(0.5)
            finally:
                self.db.disconnect()
                self.ds.disconnect()

    def process_request(self, client):
        request = client.recv()

        # A simple pre-check to check if we received a
        # valid request with "action" as keyword.
        if (not isinstance(request, dict)
                or not isinstance(request.get("action", None), str)
                or request["action"] not in [
                    "ping",
                    "server-status",
                    "server-status-extended",
                    "get-location-group",
                    "get-services",
                    "service-data"
                ]):
            client.send({"status":"err", "message":"invalid request"})
            client.disconnect()
            return

        if ("x-realip" in request
                and isinstance(request["x-realip"], str)
                and "x-authkey" in request
                and isinstance(request["x-authkey"], str)
                and "x_authkey" in self.config
                and self.config["x_authkey"] != None
                and request["x-authkey"] == self.config["x_authkey"]):
            addr = request["x-realip"]
            self.log.notice("connect from %s real ip %s" % (client.addr, addr))
            del request["x-realip"]
            del request["x-authkey"]
        else:
            addr = client.addr

        # The request will be processed.
        self.req = Request(client, addr, request)
        self.unixtime = time.time()
        self.log.set_pattern("%X", addr)

        """
        REQ (request): a request has an action and maybe data
        RES (response): the response has a status (ok, err)
                        and a message or data

        Request and response are json data.

        ping:
            REQ:    action: "ping"

            RES:    status: "ok"
                    message: "pong"

        server-status|server-status-extended:
            REQ:    action: "server-status"
                    authkey: "secret"

            RES:    status: "err"
                    message: "access denied"

            RES:    status: "ok"
                    data: {}

        get-location-group:
            REQ:    action: "get-location-group"
                    location_id: 12345
                    password: "secret"

            RES:    status: "err"
                    message: "access denied"

            RES:    status: "ok"
                    data: [12345, 12346, 12347, ...]

        get-services:
            REQ:    action: "get-services"
                    host_id: 12345
                    password: "secret"

            RES:    status: "err"
                    message: "access denied"

            RES:    status: "ok"
                    data: {}

        service-data:
            REQ:    action: "service-data"
                    host_id: 12345
                    password: "secret"
                    data: {}

            RES     no response send, connection just quits
        """

        if self.req.request["action"] == "ping":
            self.log.notice("process request %s" % request["action"])
            self.req.reply(status="ok", message="pong")
        elif self.req.request["action"] in ["server-status", "server-status-extended"]:
            self.log.notice("process request %s" % request["action"])
            self.process_server_status(self.req.request["action"])
        elif self.req.request["action"] in [
                "get-services",
                "service-data",
                "get-location-group"]:
            if self.validate_request():
                self.db.connect()
                self.ds.connect()
                if self.req.request["action"] == "get-location-group":
                    self.log.notice(
                        "process request %s location id %s agent version %s" % (
                            request["action"],
                            request["location_id"],
                            request["version"]
                        ))
                    self.process_location_group_request(
                        self.req.request["location_id"],
                        self.req.request["password"])
                elif self.check_host_auth() is True:
                    self.log.notice(
                        "process request %s host %s agent version %s" % (
                            request["action"],
                            request["host_id"],
                            request["version"]
                        ))
                    if self.req.request["action"] == "get-services":
                        self.process_get_services()
                    elif self.req.request["action"] == "service-data":
                        if "facts" in self.req.request:
                            self.update_host_facts()
                        # FOO
                        self.process_service_data()
        else:
            self.log.error("unknown action %s" % self.req.request["action"])
            self.req.reply(status="err", message="unknown action")

        self.log.info(
            "request finished (%.3fs)"
            % (time.time() - self.unixtime))

    def process_server_status(self, action):
        conf = self.config["server_status"]

        if (conf["enabled"] is False
                or ("all" not in conf["allow_from"]
                    and not netaddr.ip_in_network(
                        self.req.addr, conf["allow_from"]))
                or conf["authkey"] != self.req.request.get("authkey", None)):
            return self.req.reply(status="err", message="access denied")

        pool_stats = self.pm.process_status(simple=True)
        if action == "server-status-extended":
            return self.req.reply(status="ok", data=pool_stats)

        data = {}

        for pool in [
                "db_manager",
                "keepalived",
                "remote_scheduler",
                "timeout_scheduler",
                "wtrm_scheduler"]:
            key = "%s_total" % pool
            data[key] = pool_stats[key]

        for pool in ["wtrm_checker", "archive_stats"]:
            config_key = "pool_%s" % pool
            metric_key = "%s_total" % pool
            if self.config[config_key]["worker"] == 0:
                data[metric_key] = -1
            else:
                data[metric_key] = pool_stats[metric_key]

        for pool in ["timeout_checker", "listener"]:
            config_key = "pool_%s" % pool
            if self.config[config_key]["worker"] == 0:
                data[metric_key] = -1
            else:
                for metric in ["total", "waiting"]:
                    metric_key = "%s_%s" % (pool, metric)
                    data[metric_key] = pool_stats[metric_key]
                for metric in [
                        "processing",
                        "initiazing",
                        "sending",
                        "reading"]:
                    metric_key = "%s_%s" % (pool, metric)
                    data_key = "%s_processing" % pool
                    if data_key not in data:
                        data[data_key] = pool_stats[metric_key]
                    else:
                        data[data_key] += pool_stats[metric_key]

        if self.config["pool_remote_checker"]["worker"] == 0:
            data["remote_checker_total"] = -1
        else:
             data["remote_checker_total"] = pool_stats["remote_checker_total"]

        data["database_status"] = self._get_db_status(self.db)
        data["datastore_status"] = self._get_db_status(self.ds)
        data["queue_status"] = self._get_queue_status()
        self.req.reply(status="ok", data=data)

    def _get_queue_status(self):
        try:
            if not self.queue.ping():
                return False
        except Exception:
            return False
        return True

    def _get_db_status(self, db):
        try:
            db.connect()
            if not db.ping():
                return False
        except Exception:
            return False
        finally:
            db.disconnect()
        return True

    def validate_request(self):
        try:
            self.req.request = self.validate.agent_request(self.req.request)
        except Exception as e:
            self.log.exception()
            self.req.reply(status="err", message="invalid data")
            return None
        return True

    def process_location_group_request(self, location_id, password):
        location_group = self.db.get_location_group(location_id, password)
        if not location_group:
            self.req.reply(status="err", message="access denied")
            return

        allow_from = location_group["allow_from"]

        if (len(allow_from)
                and not netaddr.ip_in_network(self.req.addr, allow_from)):
            self.req.reply(status="err", message="access denied")
            return

        self.req.reply(
            status="ok",
            data=self.db.get_location_group_hosts(location_id))

    def check_host_auth(self):
        self.log.set_pattern("%Y", "host %s" % self.req.request["host_id"])
        self.log.info("check authorization")

        self.req.host = self.db.get_host_by_auth(
            self.req.request["host_id"],
            self.req.request["password"],
            self.req.request.get("location_id", None))

        if self.req.host is None:
            self.req.reply(status="err", message="access denied")
            return

        if not len(self.req.host["allow_from"]):
            return True

        allow_from = []
        for ip in self.req.host["allow_from"]:
            m = re.search(r"^%(IPADDR[1-4]{0,1})%\Z", ip)
            if m:
                ip = m.group(1).lower()
                if ip == "ipaddr":
                    ip = "ipaddr1"
                if self.req.host[ip] != None and len(self.req.host[ip]):
                    allow_from.append(self.req.host[ip])
            else:
                allow_from.append(ip)

        if netaddr.ip_in_network(self.req.addr, allow_from):
            return True

        self.req.reply(status="err", message="access denied")
        return

    def process_get_services(self):
        s = ProcessGetServices(db=self.db, log=self.log)
        self.req.reply(
            status="ok",
            data=s.get_local_services(host=self.req.host))

    def process_service_data(self):
        self.req.reply(status="ok")
        p = ProcessServiceData(
            db=self.db,
            ds=self.ds,
            host=self.req.host,
            config=self.config,
            unixtime=self.unixtime)
        p.run(self.req.request["data"])

    #def client_reply(self, status, message=None, data=None):
    #    response = {"status": status}
    #    if message is not None:
    #        response["message"] = message
    #    if data is not None:
    #        response["data"] = data
    #    self.req.client.send(response)
    #    self.req.client.disconnect()

    def update_host_facts(self):
        if not self.req.request["facts"]:
            return
        if self.req.host["facts"] != self.req.request["facts"]:
            self.db.update_host(self.req.host["id"], {"facts": self.req.request["facts"]})

# Each incoming agent request will be stored into a new object.
class Request:
    def __init__(self, client, addr, request):
        self.client = client
        self.addr = addr
        self.request = request
        self.host = None
        self.log = logger("bloonix")

    def reply(self, status, message=None, data=None):
        if status != "ok":
            # NEVER CHANGE THIS LOG LINE. Fail2ban.
            self.log.warning(
                "client %s %s (action=%s)"
                % (self.client.addr, message, self.request["action"]))
        response = {"status": status}
        if message is not None:
            response["message"] = message
        if data is not None:
            response["data"] = data
        self.client.send(response)
        self.client.disconnect()

    def finish(self):
        self.client.disconnect()
