import re
import time
from bloonix import utils
from bloonix import mstatus

class ProcessGetServices:
    def __init__(self, db, log):
        self.db = db
        self.log = log

    def get_remote_services(self, host=None, with_check_wtrm=None):
        self.host = host
        services = self.db.get_services(
            host_id=host["id"],
            with_vars=True,
            with_check_wtrm=with_check_wtrm,
            remote_check=1)
        return self._get_ready_services(services)

    def get_local_services(self, host):
        self.host = host
        services = self.db.get_services(
            host_id=host["id"],
            with_vars=True,
            with_check_wtrm=True,
            remote_check=0)
        return self._get_ready_services(services)

    def get_all_services(self, host):
        self.host = host
        services = services = self.db.get_services(
            host_id=host["id"],
            with_vars=True,
            with_check_wtrm=True,
            remote_check=[0,1])
        return self._get_ready_services(services)

    def get_wtrm_service(self, host, service_id):
        self.host = host
        services = self.db.get_services(service_ids=service_id)
        return self._get_ready_services(services)

    def _get_ready_services(self, services):
        ready_services = []
        host_variables = self.build_host_variables()
        poll_interval = 60

        for service_id, service in services.items():
            command_options, interval = self.build_command_options(
                service_id, service, host_variables)

            if interval and interval < poll_interval:
                poll_interval = interval

            if command_options is None:
                continue

            data = {
                "host_id": service["host_id"],
                "service_id": service["id"],
                "command": service["command"],
                "command_options": command_options,
            }

            if service["command"] == "check-cluster":
                self.add_check_cluster_status(command_options)

            if service["remote_check"] == 1:
                data["satellite_type"] = service["satellite_type"]
                data["satellites"] = service["satellites"]
                data["last_satellite"] = service["last_satellite"]

            ready_services.append(data)

        self.log.debug("send reply to agent", {
            "poll_interval": poll_interval,
            "services": ready_services})
        return {
            "poll_interval": poll_interval,
            "services": ready_services}

    def build_command_options(self, service_id, service, host_variables):
        command_options = service["command_options"]

        # updated in ProcessServiceData
        #if service["force_check"] > 0:
        #    self.db.update_service(service_id, {"force_check": 0})

        if service["enabled"] == 0:
            return None, None

        poll_interval = interval = service["interval"]
        if service["retry_interval"] < service["interval"]:
            poll_interval = service["retry_interval"]
        if service["status"] in mstatus.NOT_OK:
            if service["retry_interval"] < service["interval"]:
                interval = service["retry_interval"]

        last_check = service["last_check"]

        # 15 seconds will be subtracted from the interval. Reason:
        #
        # The poll interval of the agent and the interval of checks
        # is usually set to 60 seconds. This means that the agent should
        # correctly receive all checks with an interval of 60 seconds
        # with each poll interval. However, since the Bloonix server
        # takes a few seconds to validate and store the check data,
        # the value of last_check shifts to the point where no checks
        # are ready for the agent's next request. Example:
        #
        #   The agent polls every 60 seconds:
        #   01:00:00, 01:01:00, 01:02:00 and so on...
        #
        #   1. 00:00:00  Agent: request checks
        #   2. 00:00:01 Server: send ready checks
        #   3. 00:00:02  Agent: send data to server
        #   4. 00:00:03  Agent: sleep for 60s, next poll at 00:01:03
        #   5. 00:00:03 Server: receive data
        #   6. 00:00:05 Server: save data to db
        #                       last_check 00:00:05
        #                       next check 00:01:05
        #
        # And here is the problem between step 4 and 6. The agent polls
        # next time for checks at 00:01:03, but the server has saved the
        # data with timestamp 00:00:05. So the time has shifted by about
        # 2 seconds from the poll interval of the agent. With a check
        # interval of 60 seconds, the next check is at 00:01:05, so the
        # agent would not receive any checks. Only at the next poll
        # interval at 00:02:03 would the agent receive the checks,
        # but then almost 2 minutes have already passed. To correct
        # this difference, 15 seconds are subtracted from the interval.
        if (service["force_check"] == 0
                and last_check + interval - 15 > time.time()):
            return None, None

        if not isinstance(command_options, dict):
            return {}, poll_interval

        variables = service.get("template_variables", {})
        variables.update(host_variables)
        variables.update(service["variables"])
        for opt in command_options:
            if isinstance(command_options[opt], list):
                arr = command_options[opt]
                for i, v in enumerate(arr):
                    # web transaction workflow
                    if isinstance(v, dict):
                        for key in v:
                            if isinstance(v[key], str):
                                v[key] = self.replace_command_option_patterns(v[key], variables)
                    # string/metric condition
                    elif isinstance(v, list) and len(v) == 3 and isinstance(v[2], str):
                        v[2] = self.replace_command_option_patterns(v[2], variables)
                    elif isinstance(v, str):
                        arr[i] = self.replace_command_option_patterns(v, variables)
            elif isinstance(command_options[opt], str):
                command_options[opt] = self.replace_command_option_patterns(
                    command_options[opt], variables)

        return command_options, poll_interval

    def build_host_variables(self):
        host_variables = {
            "HOSTNAME": self.host["hostname"],
            "IPADDR": self.host["ipaddr1"],
            "IPADDR1": self.host["ipaddr1"],
            "IPADDR2": self.host["ipaddr2"],
            "IPADDR3": self.host["ipaddr3"],
            "IPADDR4": self.host["ipaddr4"]}
        host_variables.update(self.host["variables"])
        return host_variables

    def replace_command_option_patterns(self, value, variables):
        value = re.sub(
            r"%RND\(([1-9][0-9]{0,2})\)%",
            self.replace_random_pattern,
            value)

        for vkey in variables:
            if (variables[vkey] is not None
                    and re.search(r"^[a-zA-Z_0-9\.\s]+\Z", vkey)):
                value = re.sub(r"%%%s%%" % vkey, variables[vkey], value)

        return value

    def replace_random_pattern(self, m):
        return utils.pwgen(m.group(0))

    def add_check_cluster_status(self, opts):
        try:
            if ("service" not in opts
                    or not isinstance(opts["service"], list)):
                return
            services = self.db.get_service_status(opts["service"])
            for i, service_id in enumerate(opts["service"]):
                if isinstance(service_id, str) and re.search(r"^\d+\Z", service_id):
                    service_id = int(service_id)
                if isinstance(service_id, int) and service_id in services:
                    opts["service"][i] = "%s:%s" % (service_id, services[service_id])
        except Exception:
            pass
