import os
import re
import socket
import time
from bloonix.logger import logger
from bloonix.runcmd import RunCmd
from bloonix.json import json
from bloonix import timeperiod
from bloonix import mstatus
from bloonix import utils
from bloonix import smsprocessor


# Simple class to access new service configuration.
class Service:
    def __init__(
            self, service_name, service_id, status, message, debug,
            status_not_ok_since, attempt_counter, comment, description,
            highest_notification_status, notification_status, satellite_id=0):

        self.service_name = service_name
        self.service_id = service_id
        self.status = status
        self.message = message
        self.comment = comment
        self.description = description
        self.highest_notification_status = highest_notification_status
        self.notification_status = notification_status
        self.debug = debug
        # For escalation level
        self.status_not_ok_since = status_not_ok_since
        self._message = []
        self._tags = []

        # Cache new service data, which will be stored
        # to the database at the end of processing
        self.data = {"last_satellite": satellite_id, "force_check": 0}

        # Send a notification when the maximum attempts
        # have been reached or the service is flapping.
        self.attempt_counter = attempt_counter
        self.attempts_exceeded = None
        self.is_flapping = None

        # The highest fd status if the service is flapping
        # between states.
        self.highest_fd_status = None

        # Is a notification send?
        self.notification_successfully_send = None

    def get_message(self):
        if len(self._message):
            return "%s [%s]" % (self.message, "; ".join(self._message))
        return self.message

    def add_message(self, value):
        self._message.append(value)

    def get_tags(self):
        return ",".join(self._tags)

    def add_tags(self, *values):
        for value in values:
            value = value.replace(" ", "")
            self._tags += re.split(",", value)

    def set_data(self, key, value):
        self.data[key] = value

    def get_data(self):
        self.data["message"] = self.get_message()
        self.data["tags"] = self.get_tags()
        self.data["attempt_counter"] = self.attempt_counter
        return self.data


# p = ProcessServiceData(**kwargs)
# p.run(data, services)
class ProcessServiceData:
    # Cache plugin information until the plugin version
    # changes in the sys_options table. The attributes
    # are controlled in the method get_plugin_info.
    __plugin_cache = {}
    __plugin_version = None

    def __init__(self, db, ds, host, config, unixtime, update_last_check=True):
        self.db = db
        self.ds = ds
        self.host = host
        self.config = config
        self.unixtime = int(unixtime)
        # update_last_check is set to false by the
        # timeout checker, otherwise a get-services
        # request from the agent will return no services
        self.update_last_check = update_last_check
        self.data = None
        self.services = None
        self.log = logger("bloonix")
        self.plugin_version = None
        self.has_maintenance = None
        self.events = []
        self._host_is_down = None
        self.ready_for_notification = []

        # these attributes are overridden in the process() method
        # in a loop for each service to be processed
        self.cs = None # Dict, current service configuration from database
        self.ns = None # Service() object, new service data from an agent

    def run(self, data, services=None):
        # The table sys_options contains information about the current
        # plugin version and if maintenance is active. The plugin
        # version is needed to update the cached plugin information
        # in __plugin_cache as soon as the version changes.
        # plugin_version should be updated to a new version if any
        # plugin is updated in the database.
        # If maintenance is active, then notifications are suppressed.
        # In addition, events are stored with the tag "maintenance".
        # This way it is possible to filter the times from the
        # availability reports.
        opts = self.db.get_maintenance()
        self.plugin_version = opts["plugin_version"]
        self.has_maintenance = opts["has_maintenance"]

        # The services argument is set by the timeout scheduler.
        # If the argument is set, the service data does not have
        # to be selected from the database.
        data, stats, results, services = self.validate_data(data, services)

        #self.log.warning(results, dump=True)
        #self.log.warning(stats, dump=True)

        # No data, no further processing.
        if len(data) == 0:
            return

        # The first thing that is stored is the metrics.
        # If an error occurs during saving, the error is ignored
        # in order to continue processing the status of the services.
        if len(stats):
            try:
                self.ds.create_stats(stats)
            except Exception:
                self.log.exception()
                self.save_to_datastore_fallback("stat", stats)

        if len(results):
            try:
                self.ds.create_results(results)
            except Exception:
                self.log.exception()
                self.save_to_datastore_fallback("result", results)

        # Request all host and service downtimes at once.
        # Host has a active downtime?
        self.has_host_downtime = self.get_host_downtime(self.host["id"])
        if not self.has_host_downtime:
            # Processing service has a active downtime?
            # {service_id: None|True}
            self.has_service_downtime = self.get_service_downtimes(list(data))

        self.processing_service_ids = list(data)
        for service_id, service_data in data.items():
            self.process(service_id, service_data, services)
            self.update_service()

        self.create_events()
        self.send_notifications()

    def process(self, service_id, service_data, services):
        # * Maintenance active
        #    -> Skip notification
        #    -> Tag maintenance
        # * Host or service has a downtime?
        #    -> Force Status INFO
        #    -> Tag downtime
        # * Host or service is disabled?
        #    -> Force Status INFO
        #    -> Tag disabled
        # * The host alive check has a status which shows that
        #    the host has a problem
        #    -> Skip notification
        #    -> Tag host down
        # * Service is acknowledged
        #    -> Skip notification
        #    -> Tag acknowledged
        # * Attempt counter == max attempts
        #    -> Send notification
        # * Service is flapping between states
        #    -> DO NOT FORCE any status
        #    -> Force notification, simulate status CRITICAL
        # * Service recovering
        #    -> Send notification
        self.cs = services[service_id]
        self.ns = Service(
            service_name=self.cs["service_name"],
            service_id=service_id,
            status=service_data["status"],
            message=service_data["message"],
            debug=service_data.get("debug", None),
            status_not_ok_since=self.cs["status_not_ok_since"],
            attempt_counter=self.cs["attempt_counter"],
            satellite_id=service_data.get("satellite_id", 0),
            comment=self.cs["comment"],
            description=self.cs["description"],
            highest_notification_status=self.cs["highest_notification_status"],
            notification_status=self.cs["notification_status"])

        self.fix_time_issues()

        if self.update_last_check:
            self.ns.data["last_check"] = self.unixtime

        if self.has_maintenance:
            self.ns.add_tags("maintenance")

        if self.has_downtime_or_is_disabled():
            return

        if self.cs["has_downtime"] == 1:
            self.ns.data["has_downtime"] = 0

        if self.ns.status == mstatus.OK and len(self.cs["notification_status"]):
            self.ns.data["notification_status"] = []

        self.check_service_status()
        self.check_attempt_counter()
        self.check_soft_recovery()
        self.check_flap_detection()
        self.check_notification_status()
        self.check_event()

    def fix_time_issues(self):
        # NTP time issues? Fix what we can fix.
        # The following parameter cannot be higher
        # than the current unixtime.
        for k in ["last_notification", "status_since", "status_not_ok_since", "last_check"]:
            if self.cs[k] > self.unixtime:
                self.ns.data(k, self.unixtime)
                self.ns.status_not_ok_since = self.unixtime
                self.cs[k] = self.unixtime

    def update_service(self):
        data = {}
        # Update only data which has changed.
        for k, v in self.ns.get_data().items():
            if self.cs.get(k, None) != v:
                data[k] = v
        if len(data):
            self.db.update_service(self.ns.service_id, data)

    def create_events(self):
        if len(self.events):
            self.db.create_event(self.events)

    def has_downtime_or_is_disabled(self):
        if (self.host["enabled"] == 1
                and self.cs["enabled"] == 1
                and not self.has_host_downtime
                and not self.has_service_downtime[self.ns.service_id]):
            return None

        # A lot of parameter has to be reset if
        # the status is forced to INFO.
        if self.cs["acknowledged"] == 1:
            self.ns.data["acknowledged"] = 0
        if self.cs["attempt_counter"] > 0:
            self.ns.data["attempt_counter"] = 0
        if self.cs["last_notification"] > 0:
            self.ns.data["last_notification"] = 0
        if self.cs["highest_notification_status"] != mstatus.OK:
            self.ns.data["highest_notification_status"] = mstatus.OK
            self.ns.data["notification_status"] = []
        if self.cs["status_not_ok_since"] > 0:
            self.ns.data["status_not_ok_since"] = 0
            self.ns.status_not_ok_since = 0
        if self.cs["status_switches"] > 0:
            self.ns.data["status_switches"] = 0

        if self.cs["status"] != mstatus.INFO:
            self.ns.data["status"] = mstatus.INFO
            self.ns.data["status_since"] = self.unixtime

            if self.host["enabled"] == 0 or self.cs["enabled"] == 0:
                self.ns.add_tags("disabled")
            if self.has_host_downtime or self.has_service_downtime[self.ns.service_id]:
                self.ns.add_tags("downtime")
                self.ns.data["has_downtime"] = 1

            data = {
                "time": self.unixtime,
                "host_id": self.host["id"],
                "service_id": self.ns.service_id,
                "last_status": self.cs["status"],
                "status": mstatus.INFO,
                "tags": self.ns.get_tags(),
                "duration": self.unixtime - self.cs["status_since"],
                "attempts": "0/%s" % self.cs["attempt_max"],
                "message": self.ns.get_message(),
            }
            if self.ns.debug:
                data["debug"] = self.ns.debug
            self.events.append(data)

        return True

    def get_plugin_info(self, command):
        c = ProcessServiceData

        if self.plugin_version is None:
            raise Exception("something wents wrong, attribute plugin_version is empty")

        if c.__plugin_version != self.plugin_version:
            c.__plugin_cache.clear()
            c.__plugin_version = self.plugin_version

        if command not in c.__plugin_cache:
            self.log.info("get plugin information for command %s" % command)
            r = self.db.get_plugin_info(command)
            if r:
                c.__plugin_cache[command] = r
            else:
                c.__plugin_cache[command] = {}

        return c.__plugin_cache[command]

    def get_host_downtime(self, host_id):
        for row in self.db.get_host_downtimes(host_id):
            if row["timeslice"]:
                if timeperiod.check(
                        period=row["timeslice"],
                        unixtime=self.unixtime,
                        tz=row["timezone"]):
                    return True
            elif self.unixtime >= row["begin"] and self.unixtime <= row["end"]:
                return True

        return None

    def get_service_downtimes(self, service_ids):
        downtime = {}
        for service_id in service_ids:
            downtime[service_id] = None

        for row in self.db.get_service_downtimes(service_ids):
            if row["timeslice"]:
                downtime[row["service_id"]] = timeperiod.check(
                    period=row["timeslice"],
                    unixtime=self.unixtime,
                    tz=row["timezone"])
            elif self.unixtime >= row["begin"] and self.unixtime <= row["end"]:
                downtime[row["service_id"]] = True

        return downtime

    def check_service_status(self):
        # Clearing some vars if the status is ok.
        if self.ns.status == mstatus.OK:
            if self.cs["highest_notification_status"] > 0:
                self.ns.data["highest_notification_status"] = 0
                self.ns.data["notification_status"] = []
            if self.cs["acknowledged"] == 1:
                self.ns.data["acknowledged"] = 0
            if self.ns.attempt_counter > 0:
                self.ns.attempt_counter = 0
            if self.cs["status_not_ok_since"] > 0:
                self.ns.data["status_not_ok_since"] = 0
                self.ns.status_not_ok_since = 0

        if self.ns.status != self.cs["status"]:
            self.ns.data["status"] = self.ns.status
            self.ns.data["status_since"] = self.unixtime

    def check_attempt_counter(self):
        # Fixing counter
        if self.ns.attempt_counter > self.cs["attempt_max"]:
            self.ns.attempt_counter = self.cs["attempt_max"]

        if self.ns.status in mstatus.NOT_OK:
            if self.cs["status_not_ok_since"] == 0:
                self.ns.data["status_not_ok_since"] = self.unixtime
                self.ns.status_not_ok_since = self.unixtime
            if self.ns.attempt_counter < self.cs["attempt_max"]:
                self.ns.attempt_counter += 1

        if (self.ns.attempt_counter == self.cs["attempt_max"]
                or (self.ns.status == mstatus.OK
                    and self.cs["status"] in mstatus.NOT_OK)):
            self.ns.attempts_exceeded = True

    def check_soft_recovery(self):
        # After a notification has not been sent due to a dependency
        # to another service or due to a host alive check, there must
        # be a way for the services to recover, otherwise a race
        # condition can occur if the service to recover is still
        # critical, but the service to which the dependency exists
        # or the host alive check already return an OK. For this
        # purpose, the notification is suspended for an interval.
        if not self.host_is_down() and self.cs["soft_recovery"] == 1:
            self.ns.is_recovering = True
            self.ns.update("soft_recovery", 0)
            self.ns.add_tags("soft-recovery")

    def check_flap_detection(self):
        # Even if the flap detection is not switched on, the counter
        # is still incremented. If flap detection is switched on,
        # a message can be generated directly if a service is flapping.
        status_switches = self.cs["status_switches"]
        status_since = self.cs["status_since"]
        self.ns.highest_fd_status = self.cs["highest_fd_status"]

        # Fix time issues
        if self.cs["status_since"] > self.unixtime:
            status_since = self.ns.data["status_since"] = self.unixtime

        if self.ns.status == self.cs["status"]:
            # If a service switches state in the period of fd_time_range,
            # then the counter is incremented. If the status is OK in the
            # time period of fd_time_range, the counter is reset to 0.
            if self.unixtime - status_since >= self.cs["fd_time_range"]:
                status_switches = 0
                if self.cs["status_switches"] > 0:
                    self.ns.data["status_switches"] = 0
                if self.cs["highest_fd_status"] > 0:
                    self.ns.data["highest_fd_status"] = 0
        elif self.cs["status"] != mstatus.INFO:
            status_switches += 1
            self.ns.data["status_switches"] = status_switches
            if self.ns.status > self.cs["highest_fd_status"]:
                self.ns.data["highest_fd_status"] = self.ns.status
                self.ns.highest_fd_status = self.ns.status

        if (self.cs["fd_enabled"] == 1
                and status_switches >= self.cs["fd_flap_count"]):
            self.ns.add_tags("flapping")
            self.ns.add_message(
                "highest flap detection status: %s"
                % mstatus.STRING[self.cs["highest_fd_status"]])
            self.ns.is_flapping = True

        if self.ns.is_flapping and self.cs["is_flapping"] == 0:
            self.ns.data["is_flapping"] = 1
        elif not self.ns.is_flapping and self.cs["is_flapping"] == 1:
            self.ns.data["is_flapping"] = 0

    def check_event(self):
        # It's important to call this method at the end
        # of process(), otherwise tags+message could missing
        # some information.
        if (self.ns.status != self.cs["status"]
                or (self.ns.status in mstatus.NOT_OK
                    and self.ns.attempt_counter != self.cs["attempt_counter"])):
            # time issue?
            if self.cs["status_since"] > self.unixtime:
                duration = 0
            else:
                duration = self.unixtime - self.cs["status_since"]
            data = {
                "time": self.unixtime,
                "host_id": self.host["id"],
                "service_id": self.ns.service_id,
                "last_status": self.cs["status"],
                "status": self.ns.status,
                "tags": self.ns.get_tags(),
                "duration": duration,
                "attempts": "%s/%s" % (self.ns.attempt_counter, self.cs["attempt_max"]),
                "message": self.ns.get_message(),
            }
            if self.ns.debug:
                data["debug"] = self.ns.debug
            self.events.append(data)

    def get_highest_notification_status(self):
        if self.ns.attempts_exceeded is True:
            return self.ns.status

        if self.ns.is_flapping:
            return self.ns.highest_fd_status

    def host_is_down(self):
        if self._host_is_down is None:
            if self.db.host_alive_critical(self.host["id"], self.processing_service_ids):
                self._host_is_down = True
            else:
                self._host_is_down = False
        return self._host_is_down

    def check_notification_status(self):
        # Check if a notification must be send for this service.
        # If yes, then the service is saved to ready_for_notification.
        if (self.has_maintenance
                or not self.ns.attempts_exceeded
                or self.host["notification_enabled"] == 0
                or self.cs["notification_enabled"] == 0
                or self.cs["acknowledged"] == 1
                or self.cs["attempt_max"] == 0):
            self.log.info(
                "no notification:",
                "status", self.ns.status,
                "maintenance", self.has_maintenance,
                "attempts exceeded", self.ns.attempts_exceeded,
                "host notification enabled", self.host["notification_enabled"],
                "service notification enabled", self.cs["notification_enabled"],
                "acknowledged", self.cs["acknowledged"],
                "attempt_max", self.cs["attempt_max"],
                "[s=%s]" % self.ns.service_id)
            return

        if self.host_is_down():
            self.log.info(
                "no notification: host is down",
                "[s=%s]" % self.ns.service_id)
            self.ns.add_tags("host-down")
            return

        if self.ns.status in mstatus.NOT_OK:
            if self.ns.status <= self.cs["highest_notification_status"]:
                if self.cs["notification_interval"] > 0:
                    notification_interval = self.cs["notification_interval"]
                else:
                    notification_interval = self.host["notification_interval"]
                diff_time = self.unixtime - self.cs["last_notification"]
                if diff_time <= notification_interval:
                    self.log.info(
                        "no notification:",
                        "status", self.ns.status,
                        "last notification %s" % utils.timestamp(
                            self.cs["last_notification"]),
                        "interval", notification_interval,
                        "[s=%s]" % self.ns.service_id)
                    return
        elif self.cs["highest_notification_status"] == 0:
            self.log.info(
                "no notification:",
                "status", self.ns.status,
                "highest_notification_status", self.cs["highest_notification_status"],
                "[s=%s]" % self.ns.service_id)
            return

        if self.has_active_dependency():
            self.log.info(
                "no notification:",
                "status", self.ns.status,
                "dependency active",
                "[s=%s]" % self.ns.service_id)
            return

        self.ready_for_notification.append(self.ns)

    def has_active_dependency(self):
        status = self.get_highest_notification_status()
        dependencies = self.db.get_dependencies(self.ns.service_id)
        for dep in dependencies:
            if (status in dep["status"]
                    and dep["service_has_on_status"] in dep["on_status"]):
                return True
        return None

    def send_notifications(self):
        if not len(self.ready_for_notification):
            return

        service_ids = []
        for service in self.ready_for_notification:
            service_ids.append(service.service_id)

        skip_contact_ids = []
        contacts = self.db.host_contact_ms(self.host["id"])
        for contact_id, contact in contacts.items():
            skip_contact_ids.append(contact_id)

        service_contacts = self.db.service_contact_ms(service_ids, skip_contact_ids)
        contacts.update(service_contacts)
        if not len(contacts):
            return

        # Check for each contact and message service:
        # 1. check if escalation level
        # 2. check if status or highest flap detection status with
        #    the notification level
        # 3. check timeperiods

        for contact_id, contact in contacts.items():
            for message_service in contact["message_services"]:
                try:
                    self._for_each_contact_message_service(contact, message_service)
                except Exception:
                    self.log.exception(level="CRITICAL")

        service_ids_by_status = {}
        for service in self.ready_for_notification:
            if service.notification_successfully_send is not True:
                continue

            data = {"last_notification": self.unixtime}

            if (service.status in mstatus.NOT_OK
                    and service.status not in service.notification_status):
                data["notification_status"] = service.notification_status
                data["notification_status"].append(service.status)

            if service.status > service.highest_notification_status:
                data["highest_notification_status"] = service.status

            self.db.update_last_notification(service.service_id, data)

    def _for_each_contact_message_service(self, contact, message_service):
        services = self.check_service_with_contact_message_service(
            contact, message_service)
        if not len(services):
            return
        variables = self.build_message_service_variables(
            contact, services, message_service)
        params = self.build_message_service_params(
            message_service["builtin"], message_service["params"], variables)

        if message_service["builtin"] == "sendmail":
            script = "builtin-sendmail"
        elif message_service["builtin"] == "http":
            script = "builtin-http"
        elif message_service["builtin"] == "script":
            script = message_service["script"]

        if self.log.is_debug():
            self.log.debug("run message service command:", script)
            self.log.debug(params, dump=True)
        ret = self.run_message_service_script(script, params)

        error = None
        reason = None
        if ret == None:
            error = (
                "notification could not be send: "
                "contact id %s "
                "name %s "
                "send_to %s "
                "message service %s"
                % (
                    contact["id"],
                    contact["name"],
                    message_service["send_to"],
                    message_service["name"],
                ))
            reason = "message service does not exists"
        elif ret.has_failed:
            error = (
                "notification could not be send: "
                "contact id %s "
                "name %s "
                "send_to %s "
                "message service %s"
                % (
                    contact["id"],
                    contact["name"],
                    message_service["send_to"],
                    message_service["name"],
                ))
            reason = "error: %s; stderr %s; stdout %s" % (
                ret.error, ret.stderr, ret.stdout)

        if error:
            self.log.critical("%s reason: %s" % (error, reason))
            self.log.critical(script, params, dump=True)
            if len(error) > 10000:
                error = error[:10000]
            self.db.create_notification({
                "time": int(time.time()),
                "host_id": self.host["id"],
                "message_service_name": message_service["name"],
                "contact": contact["name"],
                "send_to": message_service["send_to"],
                "message": variables["message"],
                "success": 0,
                "error": reason})
            return

        self.db.create_notification({
            "time": int(time.time()),
            "host_id": self.host["id"],
            "message_service_name": message_service["name"],
            "contact": contact["name"],
            "send_to": message_service["send_to"],
            "message": variables["message"],
            "success": 1})

        for service in services:
            service.notification_successfully_send = True

    def check_service_with_contact_message_service(self, contact, message_service):
        services = []

        for service in self.ready_for_notification:
            if "services" in contact and service.service_id not in contact["services"]:
                continue

            if (contact["escalation_time"] != 0
                    and self.unixtime - service.status_not_ok_since
                        < contact["escalation_time"]):
                self.log.info(
                    "escalation level not hit [c=%s,s=%s]"
                    % (contact["id"], service.service_id))
                continue

            if service.is_flapping:
                status = service.highest_fd_status
            else:
                status = service.status

            if status not in message_service["notification_level"]:
                self.log.info(
                    "service status not in notification level [c=%s,s=%s]"
                    % (contact["id"], service.service_id))
                continue

            if status == mstatus.OK:
                hit = None
                for s in message_service["notification_level"]:
                    if s in service.notification_status:
                        hit = True
                        break
                if hit is None:
                    self.log.info(
                        "message service notifiation_level",
                        "not in service notification_status",
                        "[c=%s,s=%s]" % (contact["id"], service.service_id))
                    continue

            timeperiod_check = timeperiod.check(
                message_service["timeperiods"],
                self.unixtime,
                contact["timezone"])

            if not timeperiod_check:
                self.log.info(
                    "no active timeperiod [c=%s,s=%s]"
                    % (contact["id"], service.service_id))
                continue

            services.append(service)
        return services

    def build_message_service_variables(self, contact, services, message_service):
        v = {
            "host_id": str(self.host["id"]),
            "host_description": self.host["description"],
            "hostname": self.host["hostname"],
            "ipaddr": self.host["ipaddr1"],
            "date": utils.timestamp(
                unixtime=self.unixtime,
                timezone=contact["timezone"]),
            "unixtime": str(self.unixtime),
            "send_to": message_service["send_to"],
            "webgui_domain": self.config["webgui_domain"]}
        v["short_message"], v["sum_status"] = self.build_message_service_summary(
            services)
        v["message"] = self.build_message_service_message(
            message_service["message_template"], services, v)
        return v

    def build_message_service_summary(self, services):
        sum_status = []
        sum_status_code = {}
        service_names = []

        for service in services:
            if service.status not in sum_status_code:
                sum_status_code[service.status] = []
            sum_status_code[service.status].append(service.service_name)

        for code in reversed(mstatus.CODES):
            if code in sum_status_code:
                sum_status.append(
                    "%s=%s"
                    % (
                        mstatus.STRING[code],
                        len(sum_status_code[code])
                    ))
                service_names += sum_status_code[code]

        sum_status = " ".join(sum_status)
        service_names = ", ".join(service_names)

        if self.host_is_down():
            host_down = "HOST DOWN "
        else:
            host_down = ""

        if len(services) == 1:
            service = services[0]
            short_message = "%s%s %s %s %s" % (
                host_down,
                self.host["hostname"],
                mstatus.STRING[service.status],
                service.service_name,
                service.get_message())
        else:
            short_message = "%s%s %s %s" % (
                host_down,
                self.host["hostname"],
                sum_status,
                service_names)

        short_message = smsprocessor.truncate_to_one_sms(short_message)

        return short_message, sum_status

    def build_message_service_message(self, template, services, variables):
        template = re.sub(
            r"%%(%s)%%" % "|".join(variables.keys()),
            lambda m: variables[m.group(1)],
            template)
        m = re.search(
            r"(.*)%begin_loop%(?:\r\n|\n){0,1}(.*)%end_loop%(.*)",
            template, re.S)
        if m:
            template = "%s%s%s" % (
                m.group(1),
                self.build_message_service_loop_content(m.group(2), services),
                m.group(3))
        return template

    def build_message_service_loop_content(self, template, services):
        message = []
        for service in services:
            t = template
            t = t.replace("%service_id%", str(service.service_id))
            t = t.replace("%service_name%", service.service_name)
            t = t.replace("%service_status%", mstatus.STRING[service.status])
            t = t.replace("%service_message%", service.get_message())
            t = t.replace("%service_comment%", service.comment)
            t = t.replace("%service_description%", service.description)
            message.append(t)
        return "".join(message)

    def build_message_service_params(self, builtin, params, variables):
        _params = {}
        for key, value in params.items():
            if value == None:
                continue

            if builtin == "http" and key == "postdata_json":
                if re.search(r"\{", value):
                    json_string = re.sub(r"%%(%s)%%" % "|".join(variables.keys()),
                        lambda m: json.dumps(variables[m.group(1)])[1:][:-1],
                        value)
                    _params[key] = json.loads(json_string)
            elif builtin == "http" and (key == "header" or key == "postdata_urlencoded"):
                _params[key] = {}
                for subkey, subvalue in value.items():
                    subkey = re.sub(r"%%(%s)%%" % "|".join(variables.keys()),
                        lambda m: json.dumps(variables[m.group(1)])[1:][:-1],
                        subkey)
                    if key == "postdata_urlencoded" and isinstance(subvalue, list):
                        _params[key][subkey] = []
                        for subsubvalue in subvalue:
                            subsubvalue = re.sub(r"%%(%s)%%" % "|".join(variables.keys()),
                                lambda m: variables[m.group(1)],
                                subsubvalue)
                        _params[key][subkey].append(subsubvalue)
                    else:
                        _params[key][subkey] = re.sub(r"%%(%s)%%" % "|".join(variables.keys()),
                            lambda m: variables[m.group(1)],
                            subvalue)
            elif isinstance(value, str):
                _params[key] = re.sub(r"%%(%s)%%" % "|".join(variables.keys()),
                    lambda m: variables[m.group(1)],
                    value)
            else:
                _params[key] = value
        return _params

    def run_message_service_script(self, script, params):
        command = None
        for path in self.config["message_service_script_path"]:
            cmd = "%s/%s" % (path, script)
            if os.path.isfile(cmd):
                command = cmd
        if not command:
            self.log.error(
                "message service script %s does not exist"
                % script)
            return
        return RunCmd(
            command=[command],
            to_stdin=json.dumps(params),
            timeout=15,
            kill_signal=9)

    def validate_data(self, data, services=None):
        # data = [
        #   {
        #       "service_id": int,
        #       "data": {
        #           "status": (str, int),
        #           "message": str,
        #           "stats": dict,
        #           "result": dict | list,
        #           "debug": dict | list,
        #           "satellite_id": int,
        #       },
        #   }
        # ]

        if not isinstance(data, list):
            self.log.warning("data are not of type list [h=%s]" % self.host["id"])

        _data = {}
        stats = []
        results = []

        if not services:
            service_ids = []
            for service in data:
                if (isinstance(service, dict)
                        and "service_id" in service
                        and isinstance(service["service_id"], int)):
                    service_ids.append(service["service_id"])
            services = self.db.get_services(
                service_ids=service_ids,
                with_check_wtrm=True,
                with_vars=False)

        for item in data:
            if not self.validate_base_struct(item):
                self.log.warning("invalid data structure, host %s" % self.host["id"])
                continue

            if item["service_id"] not in services:
                self.log.warning("service %s does not exists" % item["service_id"])
                continue

            service_id = item["service_id"]
            plugin_data = item["data"]
            service = services[service_id]
            _data[service_id] = item["data"]

            if "stats" in plugin_data:
                s = self.validate_stats(
                    service, plugin_data["stats"],
                    plugin_data.get("satellite_id", None))
                if s:
                    stats += s
                else:
                    self.log.warning(
                        "invalid statistic data [h=%s,s=%s]"
                        % (self.host["id"], service_id))
                    self.log.warning(data)
                del plugin_data["stats"]

            if (service["command"] == "check-wtrm"
                    and "result" in plugin_data
                    and isinstance(plugin_data["result"], (list, dict))):
                results.append({
                    "time": self.unixtime,
                    "service_id": service_id,
                    "status": plugin_data["status"],
                    "message": plugin_data["message"],
                    "data": json.dumps(plugin_data["result"])})

            # Debug messages are only written to the events table and not
            # to the service table, because the payload would be too high
            # if debug message would be updated in the service table.
            # The events table should be sufficient for debugging.
            if ("debug" in plugin_data
                    and not isinstance(plugin_data["debug"], (dict,list))):
                self.log.warning(
                    "invalid 'debug' structure [h=%s,s=%s]"
                    % (self.host["id"], service_id))
                del plugin_data["debug"]

            if "tags" in plugin_data and not self.validate_tags(plugin_data):
                self.log.warning(
                    "invalid tags [h=%s,s=%s]"
                    % (self.host["id"], service_id))
                del plugin_data["tags"]

        return _data, stats, results, services

    def validate_base_struct(self, item):
        if (isinstance(item, dict)
                and "service_id" in item
                and isinstance(item["service_id"], int)
                and "data" in item
                and isinstance(item["data"], dict)
                and "status" in item["data"]
                and isinstance(item["data"]["status"], (str, int))
                and item["data"]["status"] in mstatus.CODE
                and "message" in item["data"]
                and isinstance(item["data"]["message"], str)):
            item["data"]["status"] = mstatus.CODE[ item["data"]["status"] ]
            return True

    def validate_tags(self, tags):
        if (isinstance(tags, str)
                and len(item["tags"]) <= 100
                and re.search(r"^[\w\-]+(,[\w\-]+){0,}\Z", tags)):
            return True

    def validate_stats(self, service, stats, satellite_id=None):
        if not isinstance(stats, dict):
            self.log.warning("invalid stats structure")
            return None

        command = service["command"]
        metrics = self.get_plugin_info(command).get("metrics", None)
        if not metrics:
            self.log.warning("command %s has no metrics configured" % command)
            return None

        return self.validate_metrics(service, metrics, stats, satellite_id)

    def validate_metrics(self, service, metrics, stats, satellite_id=None):
        data = []

        for k in list(stats):
            if k not in metrics or "item_id" not in metrics[k]:
                # If a plugin has new metrics, but they have not yet 
                # been imported into the database, then no error is
                # thrown. Instead, the metrics are deleted. This is
                # also important because plugin data is cached by
                # the server to reduce the load on the database.
                del stats[k]
                continue

            t = metrics[k]["datatype"]

            if ((t == "float" and not isinstance(stats[k], (float,int)))
                    or (t == "int" and not isinstance(stats[k], int))):
                return None

            d = {
                "time": self.unixtime,
                "service_id": service["id"],
                "item_id": metrics[k]["item_id"],
                "value": stats[k],
                "satellite_id": satellite_id,
            }
            data.append(d)

        return data

    def save_to_datastore_fallback(self, table, data):
        if self.config["datastore_fallback"]:
            f = open(self.config["datastore_fallback"], "a")
            f.write(json.dumps({"table": table, "data": data}))
            f.write("\n")
            f.close()
