import os
import time
from bloonix.logger import logger
from bloonix.dispatcher import Dispatcher
from bloonix.plugin import execute_check, execute_satellite_check

class RemoteChecker:
    def __init__(self, pm, queue, db, ds, config):
        self.pm = pm
        self.queue = queue
        self.db = db
        self.ds = ds
        self.config = config
        self.log = logger("bloonix")
        self.log.set_pattern("%Z", "remote_checker")
        pool_config = config["pool_remote_checker"]
        self.num_worker = pool_config["worker"]
        self.max_con_hosts = pool_config["max_concurrent_hosts"]
        self.max_con_checks = pool_config["max_concurrent_checks"]
        self.testmode = pool_config["testmode"]
        self.dispatcher = Dispatcher(self.num_worker, friendly_sleep_max=0.2)
        self.dispatcher.on("ready", self.ready)
        self.dispatcher.on("process", self.process)
        self.dispatcher.on("finish", self.finish)
        self.hosts_by_id = {}

    def run(self):
        self.log.debug("startup dispatcher")
        self.dispatcher.run()

    def ready(self, ready_worker):
        self.log.debug(ready_worker, "ready worker")

        if self.pm.done:
            self.dispatcher.done = True
            return

        while ready_worker > 0 and len(self.hosts_by_id) < self.max_con_hosts:
            self.log.debug("queue length =", self.queue.llen(QUEUE_READY_HOSTS))
            try: 
                self.log.debug("fetch a new host from queue")
                host_id = self.queue.pull(QUEUE_READY_HOSTS)
                if not host_id:
                    return
                host_id = int(host_id)
                if host_id in self.hosts_by_id:
                    self.log.notice("already processing host %s - ignoring" % host_id)
                    continue
                ready_worker -= 1
                self.log.notice("processing host %s" % host_id)
                self.hosts_by_id[host_id] = Host(host_id)
                self.dispatcher.add_job({
                    "action": "get-services",
                    "host_id": host_id,
                })
            except Exception:
                self.log.exception()
                time.sleep(1)

    def finish(self, success, job):
        self.log.debug("finish job", success, job)

        if not success:
            if job["action"] in ["get-services", "service-data"]:
                self.finish_host(job["host_id"])
                return

            # Should never occur
            if job["action"] != "service-check":
                raise Exception("invalid job", job)

            job = {
                "action": "service-check",
                "service_id": job["service"]["service_id"],
                "host_id": job["host_id"],
                "data": {
                    "status": "UNKNOWN",
                    "message": (
                        "agent subprocess was not successful, "
                        "see logfile for more information"
                    )
                }
            }
            self.log.error(
                "service-check failed: host %s service %s"
                % (job["host_id"], job["service_id"]))

        if job["action"] == "get-services":
            if not len(job["services"]):
                self.finish_host(job["host_id"])
                return
            host = self.hosts_by_id[job["host_id"]]
            host.services = job["services"]
            host.num_services = len(host.services)
            host.services_in_progress = 0
            while (len(host.services)
                    and host.services_in_progress
                        < self.max_con_checks):
                host.services_in_progress += 1
                self.dispatcher.add_job({
                    "action": "service-check",
                    "service": host.services.pop(),
                    "host_id": job["host_id"]
                })
        elif job["action"] in ["service-data", "delete-host"]:
            self.finish_host(job["host_id"])
        elif job["action"] == "service-check":
            host = self.hosts_by_id[job["host_id"]]
            host.services_finished.append({
                "service_id": job["service_id"],
                "data": job["data"],
            })
            if len(host.services_finished) == host.num_services:
                self.dispatcher.add_job({
                    "action": "service-data",
                    "host_id": host.host_id,
                    "data": host.services_finished,
                })
            if len(host.services):
                self.dispatcher.add_job({
                    "action": "service-check",
                    "service": host.services.pop(),
                    "host_id": job["host_id"]
                })

    def finish_host(self, host_id):
        host = self.hosts_by_id[host_id]
        self.log.notice("finish host %s" % host_id)
        del self.hosts_by_id[host_id]
        try:
            self.queue.push(QUEUE_FINISHED_HOSTS, host.host_id)
        except Exception:
            self.log.exception()

    def process(self, job):
        self.log.debug("process job", job)
        try:
            self.db.connect()
            return self._process(job)
        finally:
            self.db.disconnect()
            self.ds.disconnect()

    def _process(self, job):
        if job["action"] in ["get-services", "service-data"]:
            host = self.db.get_host(job["host_id"])
            if not host:
                return {
                    "action": "delete-host",
                    "host_id": job["host_id"]
                }

            if job["action"] == "get-services":
                if self.testmode == True:
                    return {
                        "action": "get-services",
                        "services": [],
                        "host_id": job["host_id"],
                    }
                s = ProcessGetServices(
                    db=self.db,
                    log=self.log)
                data = s.get_remote_services(host=host) #, with_check_wtrm=True)
                return {
                    "action": "get-services",
                    "services": data["services"],
                    "host_id": job["host_id"],
                }

            self.ds.connect()
            if job["action"] == "service-data":
                p = ProcessServiceData(
                    db=self.db,
                    ds=self.ds,
                    host=host,
                    config=self.config,
                    unixtime=time.time())
                p.run(job["data"])
                return {
                    "action": "service-data",
                    "host_id": job["host_id"]
                }

        if job["action"] == "service-check":
            service = job["service"]
            self.log.notice(
                "check host %s service %s command %s"
                % (service["host_id"], service["service_id"], service["command"]))

            if service["satellite_type"] == 0:
                data = execute_check(
                    command=service["command"],
                    command_options=service.get("command_options", None),
                    plugin_path=self.config["plugin_path"],
                    timeout=60,
                    host_id=service["host_id"],
                    service_id=service["service_id"])
            else:
                # The connect timeout to the satellites is 10s.
                # If a satellite is not reachable, the system jumps to the next one.
                # If 3 are not reachable, a maximum of 30s has passed.
                # But if 2 are not reachable (connect timeout 10 * 2 = 20s)
                # and the third takes 70s (60s for the check itself, 10s overhead),
                # then it can take up to 90s. We also set another 10s as overhead
                # here, which results in a timeout of 100s.
                data = execute_satellite_check(
                    command=service["command"],
                    command_options=service.get("command_options", None),
                    timeout=100,
                    last_satellite=service["last_satellite"],
                    satellites=service["satellites"],
                    satellites_by_id=self.db.get_satellites(
                        service["satellites"],
                        self.config["satellites"]
                    ),
                    host_id=service["host_id"],
                    service_id=service["service_id"])

            self.log.notice(
                "finished check host %s service %s command %s"
                % (service["host_id"], service["service_id"], service["command"]))

            return {
                "action": "service-check",
                "service_id": service["service_id"],
                "data": data,
                "host_id": job["host_id"]
            }

        # Should never occur
        raise Exception("invalid job", job)


class Host:
    def __init__(self, host_id):
        self.host_id = host_id
        self.services = None
        self.num_services = 0
        self.in_progress = 0
        self.services_in_progress = 0
        self.services_finished = []
