import re
import time
from bloonix.logger import logger

class RemoteScheduler(CheckPrimary):
    def __init__(self, pm, queue, db, me):
        self.pm = pm
        self.queue = queue
        self.db = db
        self.me = me
        self.last_host_update = 0
        self.hosts = {}
        self.queue_uptime = 0
        self.primary = None
        self.log = logger("bloonix")
        self.log.set_pattern("%Z", "remote_scheduler")

    def run(self):
        self.log.notice("up and running")
        while self.pm.done is False:
            try:
                self.db.connect()
                if self.check_me():
                    self.update_hosts()
                    self.manage_queue()
            except Exception:
                self.log.exception()
                time.sleep(1)
            finally:
                self.db.disconnect()
            time.sleep(0.5)

    def update_hosts(self):
        # Time issues?
        if self.last_host_update > time.time():
            self.last_host_update = time.time()
            return

        if self.last_host_update + 60 > time.time():
            return

        self.last_host_update = time.time()
        self.log.notice("fetch remote services")
        services = self.db.get_remote_services_without_wtrm()
        self.log.notice("%s services fetched" % len(services))
        hosts = {}

        if len(self.hosts) == 0:
            init = True
        else:
            init = False

        for service in services:
            host_id = service["host_id"]
            if host_id not in hosts:
                hosts[host_id] = None
            if host_id not in self.hosts:
                self.hosts[host_id] = {
                    "last_check": 0,
                    "queued": False,
                }

            host = self.hosts[host_id]

            if init and service["last_check"] > host["last_check"]:
                host["last_check"] = service["last_check"]

            interval = service["interval"]

            # Force check must be taken into account
            if service["retry_interval"] < 60:
                host["poll_interval"] = service["retry_interval"]
            elif service["interval"] < 60:
                host["poll_interval"] = service["interval"]
            else:
                host["poll_interval"] = 60

        for host_id in list(self.hosts):
            if host_id not in hosts:
                del self.hosts[host_id]

    def manage_queue(self):
        while True:
            self.log.debug("read from queue")
            host_id = self.queue.pull(QUEUE_FINISHED_HOSTS)
            if not host_id or not re.search(r"^\d+\Z", host_id):
                break
            host_id = int(host_id)
            if host_id in self.hosts:
                self.log.notice("host %s finished" % host_id)
                last_check = int(time.time())
                self.hosts[host_id]["last_check"] = last_check
                self.hosts[host_id]["queued"] = False

        for host_id, host in self.hosts.items():
            # Timed out? Server dead?
            if self.hosts[host_id]["queued"]:
                if time.time() - self.hosts[host_id]["queued"] > SERVICE_QUEUE_TIMEOUT:
                    queued_hosts = self.queue.get_range(QUEUE_READY_HOSTS)
                    if str(host_id) in queued_hosts:
                        # the host is in queue and waiting to get
                        # get pulled out from a remote_checker.
                        # max_concurrent_hosts should be increased
                        self.log.alert(
                            "host %s still in queue over %ss, "
                            "maybe increasing workers, max_concurrent_hosts "
                            "and max_concurrent_checks for pool remote_checker"
                            % (host_id, SERVICE_QUEUE_TIMEOUT))
                        self.hosts[host_id]["queued"] = time.time()
                    else:
                        # the host is not in queue, but in progress
                        # by a remote_checker. max_concurrent_checks
                        # should be increased
                        self.log.alert(
                            "host %s still in progress over %ss, "
                            "maybe increasing workers, max_concurrent_hosts "
                            "and max_concurrent_checks for pool remote_checker"
                            % (host_id, SERVICE_QUEUE_TIMEOUT))
                        self.hosts[host_id]["queued"] = False
                        self.hosts[host_id]["last_check"] = 0

            if (not self.hosts[host_id]["queued"]
                    and time.time()
                        - host["last_check"] > host["poll_interval"]):
                self.log.notice("push host %s into queue" % host_id)
                self.queue.push(QUEUE_READY_HOSTS, host_id)
                self.hosts[host_id]["queued"] = time.time()

    def check_me(self):
        if self.am_i_primary():
            if self.primary != self.me or self.queue_uptime > self.queue.uptime():
                self.primary = self.me
                self.last_host_update = 0
                self.hosts.clear()
                self.log.notice("flush queue QUEUE_READY_HOSTS")
                self.queue.delete(QUEUE_READY_HOSTS)
                self.log.notice("flush queue QUEUE_FINISHED_HOSTS")
                self.queue.delete(QUEUE_FINISHED_HOSTS)
            self.queue_uptime = self.queue.uptime()
            return True

        self.hosts.clear()
        self.primary = None
        self.queue_uptime = self.queue.uptime()
        return None
