import os
import setproctitle
import sys
import socket
import time
import uuid

from bloonix.logger import logger
from bloonix.procmanager import ProcManager
from bloonix.queue import Queue
from bloonix import procutils
from bloonix import mstatus
import bloonix.socket.serialized as INET

logger("bloonix").set_pattern("%X", "-") # client ip
logger("bloonix").set_pattern("%Y", "-") # host id
logger("bloonix").set_pattern("%Z", "main") # parent or pool name

setproctitle.setproctitle("bloonix-server: main")

class Server:
    VERSION = "1.35"

    def __init__(
            self, config_file, pid_file,
            background=True, log_level="debug"):
        self.config_file = config_file
        self.pid_file = pid_file
        self.validate = Validate()
        self.config = self.validate.config_file(config_file)
        self.me = "%s:%s" % (uuid.uuid4(), socket.gethostname())
        # Init listener before the user is switched, because the private key
        # for ssl encryption is maybe only accessable by root. And listen
        # ports lower than 1024 can only be initiated by root.
        self.sock = INET.Server(**self.config["listener"])

        if not os.path.isdir(self.config["plugin_cache_dir"]):
            os.makedirs(self.config["plugin_cache_dir"], mode=0o755)
            if self.config["user"] is not None:
                os.chown(
                    self.config["plugin_cache_dir"],
                    procutils.get_uid(self.config["user"]),
                    procutils.get_gid(self.config["group"])
                )

        # Switch user and group is necessary. Actions that need to
        # executed as root should done before.
        if background is True:
            procutils.hang_up(
                self.config["user"],
                self.config["group"],
                self.pid_file)
        else:
            procutils.switch_to(
                self.config["user"],
                self.config["group"])
            self.config["logger"] = {
                "screen": {
                    "level": log_level,
                    "message_layout": "%T %L %P %t %m %X %Y %Z",
                    "timeformat": "%H:%M:%S"
                }
            }

        # Load the application wide logger "bloonix"
        self.log = logger("bloonix")
        self.log.config(self.config["logger"])

        # Create the socket at first. If the socket
        # couldn't be created then it seems that another
        # bloonix server is running and holding the port.
        self.sock.listen()

        self.queue = Queue.setup(**self.config["queue"])
        self.db = Database(**self.config["database"])
        self.ds = Datastore(**self.config["datastore"])
        self.pm = ProcManager()
        self.log.notice("bloonix-server version %s started" % self.VERSION)
        self.log.notice(
            "available compression methods:",
            INET.Server.get_available_compression_methods())
        self.log.notice(
            "allowed compression methods:",
            self.sock.get_allowed_compression_methods())

        # At first start the database manager, because the
        # process creates important table partitions.
        self.pm.add_pool(
            name="db_manager",
            worker=1,
            callback=self.run_db_manager)
        self.pm.add_pool(
            name="keepalived",
            worker=1,
            callback=self.run_keepalived,
            exit_on_signal_term=False)
        self.pm.add_pool(
            name="remote_scheduler",
            worker=1,
            callback=self.run_remote_scheduler)
        if self.config["pool_remote_checker"]["worker"] > 0:
            self.pm.add_pool(
                name="remote_checker",
                worker=1,
                callback=self.run_remote_checker)
        self.pm.add_pool(
            name="timeout_scheduler",
            worker=1,
            callback=self.run_timeout_scheduler)
        self.pm.add_pool(
            name="timeout_checker",
            worker=self.config["pool_timeout_checker"]["worker"],
            callback=self.run_timeout_checker)
        if self.config["pool_listener"]["worker"] > 0:
            self.pm.add_pool(
                name="listener",
                worker=self.config["pool_listener"]["worker"],
                callback=self.run_listener)
        self.pm.add_pool(
            name="wtrm_scheduler",
            worker=1,
            callback=self.run_wtrm_scheduler)
        if self.config["pool_wtrm_checker"]["worker"] > 0:
            self.pm.add_pool(
                name="wtrm_checker",
                worker=self.config["pool_wtrm_checker"]["worker"],
                callback=self.run_wtrm_checker)
        if self.config["pool_archive_stats"]["worker"] > 0:
            self.pm.add_pool(
                name="archive_stats",
                worker=self.config["pool_archive_stats"]["worker"],
                callback=self.run_archive_stats)

    def startup(self):
        self.log.notice("startup server")
        self.run_maintenance()
        self.pm.daemonize()

    def run_listener(self, pm):
        setproctitle.setproctitle("bloonix-server: listener")
        self.log.set_pattern("%Z", "listener")
        pool = Listener(
            pm=pm,
            sock=self.sock,
            queue=self.queue,
            db=self.db,
            ds=self.ds,
            config=self.config)
        pool.run()

    def run_remote_scheduler(self, pm):
        setproctitle.setproctitle("bloonix-server: remote_scheduler")
        self.log.set_pattern("%Z", "remote_scheduler")
        pool = RemoteScheduler(
            pm=pm,
            queue=self.queue,
            db=self.db,
            me=self.me)
        pool.run()

    def run_remote_checker(self, pm):
        setproctitle.setproctitle("bloonix-server: remote_checker")
        self.log.set_pattern("%Z", "remote_checker")
        pool = RemoteChecker(
            pm=pm,
            queue=self.queue,
            db=self.db,
            ds=self.ds,
            config=self.config)
        pool.run()

    def run_keepalived(self, pm):
        setproctitle.setproctitle("bloonix-server: keepalived")
        self.log.set_pattern("%Z", "keepalived")
        pool = Keepalived(
            pm=pm,
            queue=self.queue,
            me=self.me,
            db=self.db)
        pool.run()

    def run_db_manager(self, pm):
        setproctitle.setproctitle("bloonix-server: db_manager")
        self.log.set_pattern("%Z", "db_manager")
        pool = DBManager(
            pm=pm,
            queue=self.queue,
            db=self.db,
            ds=self.ds,
            config=self.config,
            me=self.me)
        pool.run()

    def run_archive_stats(self, pm):
        setproctitle.setproctitle("bloonix-server: archive_stats")
        self.log.set_pattern("%Z", "archive_stats")
        pool = ArchiveStats(
            pm=pm,
            queue=self.queue,
            db=self.db,
            ds=self.ds,
            config=self.config)
        pool.run()

    def run_timeout_scheduler(self, pm):
        setproctitle.setproctitle("bloonix-server: timeout_scheduler")
        self.log.set_pattern("%Z", "timeout_scheduler")
        pool = TimeoutScheduler(
            pm=pm,
            queue=self.queue,
            db=self.db,
            me=self.me)
        pool.run()

    def run_timeout_checker(self, pm):
        setproctitle.setproctitle("bloonix-server: timeout_checker")
        self.log.set_pattern("%Z", "timeout_checker")
        pool = TimeoutChecker(
            pm=pm,
            queue=self.queue,
            db=self.db,
            ds=self.ds,
            config=self.config)
        pool.run()

    def run_wtrm_scheduler(self, pm):
        setproctitle.setproctitle("bloonix-server: wtrm_scheduler")
        self.log.set_pattern("%Z", "wtrm_scheduler")
        pool = WebTransactionScheduler(
            pm=pm,
            queue=self.queue,
            db=self.db,
            me=self.me)
        pool.run()

    def run_wtrm_checker(self, pm):
        setproctitle.setproctitle("bloonix-server: wtrm_checker")
        self.log.set_pattern("%Z", "wtrm_checker")
        pool = WebTransactionChecker(
            pm=pm,
            queue=self.queue,
            db=self.db,
            ds=self.ds,
            config=self.config)
        pool.run()

    def run_maintenance(self):
        # Some database changes only affect the server, not the WebGUI.
        # Any schema updates that don't affect the WebGUI can be
        # performed here. It's important that the model_version
        # remains unchanged.
        self.log.notice("maintenance start")
        db_lock = Database(**self.config["database"])
        db_lock.dbi.autocommit_default(False)
        lock_table = {"db_manager": "write"}
        done = False
        while done is not True:
            try:
                self.db.connect()
                self.ds.connect()
                db_lock.dbi.connect()
                db_lock.dbi.lock(lock_table)
                self._run_maintenance()
                done = True
            except Exception:
                self.log.exception()
            try:
                db_lock.dbi.commit()
                db_lock.dbi.unlock(lock_table)
            except Exception:
                self.log.exception()
            db_lock.disconnect()
            self.db.disconnect()
            self.ds.disconnect()
            if done is not True: time.sleep(5)

        self.log.notice("maintenance finished")

    def _run_maintenance(self):
        self.__m_service()
        self.__m_archive()

    def __m_service(self):
        if self.db.column_exists("service", "notification_status"):
            return

        if self.config["database"]["driver"] == "pg":
            self.db.do(
                'alter table "service" add column '
                '"notification_status" json not null default %s',
                ["[]"])
        else:
            self.db.do(
                'alter table `service` add column '
                '`notification_status` json not null default %s',
                ["[]"])
        for status in [2,3,4,5,6]:
            stmt, bind = self.db.sql.update(
                table="service",
                data={"notification_status": [status]},
                condition=["where", ["highest_notification_status", status]])
            self.db.do(stmt, bind)

    def __m_archive(self):
        if self.ds.table_exist("stat_hourly"):
            return
            
        if self.config["datastore"]["driver"] == "pg":
            self.ds.do(
                'create table "stat_hourly" ('
                '  "time" bigint not null,'
                '  "service_id" bigint not null,'
                '  "satellite_id" bigint null,'
                '  "item_id" smallint not null,'
                '  "avg" double precision null,'
                '  "min" double precision null,'
                '  "max" double precision null'
                ') PARTITION BY RANGE ("time")')
            self.ds.do(
                'create index "stat_hourly_time_index" on "stat_hourly" ("time")')
            self.ds.do(
                'create index "stat_hourly_service_id_time_item_id_index" on "stat_hourly" '
                '("service_id", "time", "item_id")')
        else:
            self.ds.do(
                'create table `stat_hourly` ('
                '  `time` bigint not null,'
                '  `service_id` bigint not null,'
                '  `satellite_id` bigint null,'
                '  `item_id` smallint not null,'
                '  `avg` double null,'
                '  `min` double null,'
                '  `max` double null'
                ') engine=innodb '
                'PARTITION BY RANGE (`time`) (partition p0 values less than (0))')
            self.ds.do(
                'create index `stat_hourly_time_index` on `stat_hourly` (`time`)')
            self.ds.do(
                'create index `stat_hourly_service_id_time_item_id_index` on `stat_hourly` '
                '(`service_id`, `time`, `item_id`)')

