import time
from bloonix.logger import logger

class TimeoutScheduler(CheckPrimary):
    def __init__(self, pm, queue, db, me):
        self.pm = pm
        self.queue = queue
        self.db = db
        self.me = me
        self.log = logger("bloonix")
        self.log.set_pattern("%Z", "timeout_scheduler")
        self.primary = None
        # seen_hosts is used to push hosts which have been
        # pushed into the queue or are still in the queue,
        # only after 60 seconds into the queue again.
        self.seen_hosts = {}
        self.wait_before_start = 180

    def run(self):
        # Fresh restart? Reboot?
        # If the server has been offline for a long time, then of course
        # many checks are overdue. In order to prevent the TimeoutChecker
        # from immediately issuing critical messages and to give the agents
        # the opportunity to deliver service data, it is first necessary
        # to wait for a short time.
        self.log.notice("up and running")
        self.log.notice("wait for %s seconds before start" % self.wait_before_start)
        time.sleep(self.wait_before_start)

        while self.pm.done is False:
            try:
                self.db.connect()
                if self.am_i_primary():
                    self.check_services()
                elif len(self.seen_hosts):
                    self.seen_hosts.clear()
            except Exception:
                self.log.exception()
            finally:
                self.db.disconnect()
            time.sleep(30)

    def check_services(self):
        self.log.notice("check timed out hosts")

        hosts = {}
        services = self.db.get_timed_out_services(int(time.time()) - SERVICE_TIMEOUT)
        fix_time = []
        for service in services:
            # interval_changed_at is used to check whether the interval
            # of a service has been changed recently, as it may be that
            # the timeout_scheduler recognizes a service as expired if
            # the interval has been set smaller
            if service["interval_changed_at"] > time.time():
                fix_time.append(service["service_parameter_id"])
            elif service["interval_changed_at"] + SERVICE_TIMEOUT > time.time():
                continue
            hosts[service["host_id"]] = None

        if len(fix_time):
            self.log.warning("fix interval_changed_at for multiple services")
            self.db.fix_interval_changed_at(fix_time, int(time.time()))

        self.log.notice("found %s hosts" % len(hosts))
        self.log.notice("get queued timed out hosts")
        queued_hosts = self.queue.get_range(QUEUE_TIMED_OUT_HOSTS)

        # update seen_hosts time stamp
        for host_id in queued_hosts:
            self.seen_hosts[int(host_id)] = time.time()

        self.log.notice("%s hosts in timeout queue" % len(queued_hosts))

        for host_id in list(self.seen_hosts):
            if self.seen_hosts[host_id] + 60 < time.time():
                self.log.info("remove host %s from seen hosts" % host_id)
                del self.seen_hosts[host_id]
            elif host_id in hosts:
                self.log.info("host %s already queued in the last 60 seconds" % host_id)
                del hosts[host_id]

        self.log.notice("push %s hosts into queue" % len(hosts))
        for host_id in hosts:
            self.log.info("push host %s into queue" % host_id)
            self.queue.push(QUEUE_TIMED_OUT_HOSTS, host_id)
            self.seen_hosts[host_id] = time.time()
