import os
import re
import socket
from bloonix.config import parse_file
from bloonix.validator import Validator

class Validate:
    def __init__(self):
        self.init_main()
        self.init_server_status()
        self.init_pool_listener()
        self.init_pool_remote_checker()
        self.init_pool_timeout_checker()
        self.init_pool_wtrm_checker()
        self.init_pool_archive_stats()
        self.init_agent_request()

    def config_file(self, config_file):
        config = self.v_main.validate(parse_file(config_file))

        for k, v in config["environ"].items():
            if (isinstance(k, str)
                    and isinstance(v, str)
                    and re.search(r"^[A-Z0-9_]+", k)):
                os.environ[k] = v

        # Currently not in use.
        #if config["hostname"] == "auto":
        #    config["hostname"] = socket.gethostname()

        if config["webgui_domain"] == "auto":
            config["webgui_domain"] = socket.gethostname()

        # deprecated configuration
        if "server" in config and len(config["server"]):
            config["listener"] = config["server"]
            del config["server"]

        if "bind" not in config["listener"]:
            config["listener"]["bind"] = "::"

        config["server_status"] = self.server_status(config["server_status"])
        config["pool_listener"] = self.pool_listener(config["pool_listener"])
        config["pool_timeout_checker"] = self.pool_timeout_checker(config["pool_timeout_checker"])
        config["pool_remote_checker"] = self.pool_remote_checker(config["pool_remote_checker"])
        config["pool_wtrm_checker"] = self.pool_wtrm_checker(config["pool_wtrm_checker"])
        config["pool_archive_stats"] = self.pool_archive_stats(config["pool_archive_stats"])

        config["message_service_script_path"] = re.sub(
            r"[\s\n]", "", config["message_service_script_path"])
        config["message_service_script_path"] = re.split(
            ",", config["message_service_script_path"])

        for key in ["plugin_path", "simple_plugin_path"]:
            value = config[key]
            value = re.sub(r"[\s\n]", "", value)
            value = re.split(",", value)
            config[key] = value

        return config

    def init_main(self):
        self.v_main = Validator()
        self.v_main.set({
            "webgui_domain": {
                "type": str,
                "default": "auto",
            },
            "user": {
                "type": str,
                "default": "bloonix"
            },
            "group": {
                "type": str,
                "default": "bloonix"
            },
            #"hostname": {
            #    "type": str,
            #    "default": "auto"
            #},
            #"timezone": {
            #    "type": str,
            #    "optional": True
            #},
            "message_service_script_path": {
                "type": str,
                "default": (
                    "/usr/lib/bloonix/message-service,"
                    "/usr/local/lib/bloonix/message-service")
            },
            "listener": {
                "type": dict,
                "allow_empty": True,
                "default": {}
            },
            # deprecated
            "server": {
                "type": dict,
                "allow_empty": True,
                "default": {}
            },
            "cache": {
                "type": dict,
                "allow_empty": True,
                "default": None,
                #"default": {
                #    "driver": "memcached",
                #    "host": "127.0.0.1",
                #    "port": 11211
                #}
            },
            "queue": {
                "type": dict,
                "allow_empty": True,
                "default": {
                    "driver": "redis",
                    "host": "127.0.0.1",
                    "port": 6379
                }
            },
            "pool_listener": {
                "type": dict,
                "default": {},
            },
            "pool_remote_checker": {
                "type": dict,
                "default": {},
            },
            "pool_timeout_checker": {
                "type": dict,
                "default": {},
            },
            "pool_wtrm_checker": {
                "type": dict,
                "default": {},
            },
            "pool_archive_stats": {
                "type": dict,
                "default": {},
            },
            "plugin_path": {
                "type": str,
                "default": (
                    "/usr/lib/bloonix/plugins,"
                    "/usr/local/lib/bloonix/plugins")
            },
            "simple_plugin_path": {
                "type": str,
                "default": (
                    "/usr/lib/bloonix/simple-plugins"
                    "/usr/local/lib/bloonix/simple-plugins")
            },
            "plugin_cache_dir": {
                "type": str,
                "default": "/var/lib/bloonix/plugins"
            },
            "server_status": {
                "type": dict,
                "allow_empty": True,
                "default": {}
            },
            "database": {
                "type": dict
            },
            "datastore": {
                "type": dict
            },
            "datastore_fallback": {
                "type": str,
                "optional": True,
            },
            "logger": {
                "type": dict,
                "allow_empty": True,
                "default": {}
            },
            "satellites": {
                "type": dict,
                "allow_empty": True,
                "default": {}
            },
            "environ": {
                "type": dict,
                "allow_empty": True,
                "default": {}
            },
            "x_authkey": {
                "type": str,
                "maxlength": 128,
                "optional": True,
            },
        })

    def init_server_status(self):
        self.v_server_status = Validator()
        self.v_server_status.set({
            "enabled": {
                "type": bool,
                "default": False,
            },
            "allow_from": {
                "type": str,
                "default": "127.0.0.0/8,::1/128"
            },
            "authkey": {
                "type": str,
                "optional": True
            }
        })

    def server_status(self, config):
        config = self.v_server_status.validate(config)
        config["allow_from"] = re.split(r"\s*,\s*", config["allow_from"])
        return config

    def init_pool_listener(self):
        self.v_pool_listener = Validator()
        self.v_pool_listener.set({
            "worker": {
                "type": int,
                "default": 10
            }
        })

    def pool_listener(self, config):
        return self.v_pool_listener.validate(config)

    def init_pool_remote_checker(self):
        self.v_pool_remote_checker = Validator()
        self.v_pool_remote_checker.set({
            "worker": {
                "type": int,
                "default": 10
            },
            "max_concurrent_hosts": {
                "type": int,
                "default": 0
            },
            "max_concurrent_checks": {
                "type": int,
                "default": 0
            },
            "testmode": {
                "type": bool,
                "default": False
            },
        })

    def pool_remote_checker(self, config):
        config = self.v_pool_remote_checker.validate(config)
        if config["max_concurrent_hosts"] == 0:
            config["max_concurrent_hosts"] = int(config["worker"] / 2)
        if config["max_concurrent_checks"] == 0:
            if config["worker"] < 10:
                config["max_concurrent_checks"] = config["worker"]
            else:
                config["max_concurrent_checks"] = 10
        return config

    def init_pool_timeout_checker(self):
        self.v_pool_timeout_checker = Validator()
        self.v_pool_timeout_checker.set({
            "worker": {
                "type": int,
                "default": 10
            }
        })

    def pool_timeout_checker(self, config):
        return self.v_pool_timeout_checker.validate(config)

    def init_pool_wtrm_checker(self):
        self.v_pool_wtrm_checker = Validator()
        self.v_pool_wtrm_checker.set({
            "worker": {
                "type": int,
                "default": 4,
            },
        })

    def pool_wtrm_checker(self, config):
        return self.v_pool_wtrm_checker.validate(config)

    def init_pool_archive_stats(self):
        self.v_pool_archive_stats = Validator()
        self.v_pool_archive_stats.set({
            "worker": {
                "type": int,
                "default": 2,
            },
        })

    def pool_archive_stats(self, config):
        return self.v_pool_archive_stats.validate(config)

    def init_agent_request(self):
        self.v_get_services = self.base_agent_options()
        self.v_service_data = self.base_agent_options()
        self.v_get_location_group = self.base_agent_options()
        self.v_get_services.set({
            "host_id": {
                "type": str,
                "regex": r"^[a-z0-9\-\.]+\Z",
                "default": 0
            },
            # bloonix-proxy, ignored
            "proxy": {
                "type": str,
                "default": "none"
            },
        })
        self.v_service_data.set({
            "data": {
                "type": list
            },
            "host_id": {
                "type": str,
                "regex": r"^[a-z0-9\-\.]+\Z",
                "default": 0
            },
            "facts": {
                "type": dict,
                "optional": True,
                "function": self.validate_facts,
            },
            # bloonix-proxy, ignored
            "proxy": {
                "type": str,
                "default": "none"
            },
        })
        self.v_get_location_group.set({
            "location_id": {
                "type": int
            },
            # bloonix-proxy, ignored
            "proxy": {
                "type": str,
                "default": "none"
            },
        })

    def validate_facts(self, value):
        """
        {
          "System": "Linux",
          "Hostname": "js.bloonix.dev",
          "Release": "6.1.0-11-amd64",
          "Version": "#1 SMP PREEMPT_DYNAMIC Debian 6.1.38-4 (2023-08-08)",
          "Machine": "x86_64",
          "Distrubution": {
            "Distributor ID": "Debian",
            "Description": "Debian GNU/Linux 12 (bookworm)",
            "Release": "12",
            "Codename": "bookworm"
          },
          "Memory": 33653166080,
          "Swap": 2046816256,
          "Processors": 16,
          "Interfaces": [
            "lo",
            "eth0",
            "docker0"
          ],
          "Disks": [
            "vda"
          ]
        }
        """
        if value is None:
            return True
        for key in value:
            if key not in [
                    "System", "Hostname", "Release", "Version", "Machine",
                    "Distrubution", "Memory", "Swap", "Processors",
                    "Interfaces", "Disks"]:
                del value[key]
        for key in ["System", "Hostname", "Release", "Version", "Machine"]:
            if not isinstance(value[key], str):
                del value[key]
        for key in ["Memory", "Swap", "Processors"]:
            if key in value and not isinstance(value[key], int):
                del value[key]
        for key in ["Interfaces", "Disks"]:
            if key in value and not isinstance(value[key], list):
                del value[key]
            for d in list(value[key]):
                if not isinstance(d, str):
                    del value[key]
                    break
        if "Distrubution" in value:
            if not isinstance(value["Distrubution"], dict):
                del value["Distrubution"]
            else:
                for k in value["Distrubution"]:
                    if not isinstance(value["Distrubution"][k], str):
                        del value["Distrubution"][k]
        return True

    def base_agent_options(self):
        v = Validator()
        v.set({
            "action": {
                "type": str,
                "regex": r"^(?:get-services|service-data|get-location-group)\Z"
            },
            "version": {
                "type": str,
                "regex": r"^\d+(?:\.\d+){0,1}\Z",
                "default": "0"
            },
            "password": {
                "type": str
            },
        })
        return v

    def agent_request(self, req):
        if req["action"] == "get-services":
            return self.v_get_services.validate(req)
        if req["action"] == "service-data":
            return self.v_service_data.validate(req)
        if req["action"] == "get-location-group":
            return self.v_get_location_group.validate(req)
        raise Exception("an unxpected error occurs")
